"""
****************************************************************************************************************************************************************
****************************************************************************************************************************************************************

    ___    ____  __  ___            ____  __  _______     ____               _           __
   /   |  / __ \/  |/  /           / __ \/  |/  / __ \   / __ \_________    (_)__  _____/ /_
  / /| | / / / / /|_/ /  ______   / /_/ / /|_/ / / / /  / /_/ / ___/ __ \  / / _ \/ ___/ __/
 / ___ |/ /_/ / /  / /  /_____/  / ____/ /  / / /_/ /  / ____/ /  / /_/ / / /  __/ /__/ /_
/_/  |_/_____/_/  /_/           /_/   /_/  /_/_____/  /_/   /_/   \____/_/ /\___/\___/\__/
                                                                      /___/

****************************************************************************************************************************************************************
****************************************************************************************************************************************************************

Copyright (c) 2018, Dolby Laboratories Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

****************************************************************************************************************************************************************
****************************************************************************************************************************************************************
"""

# ************************************************************************************************************************************************************ #
# ************************************************************************************************************************************************************ #


class ProfessionalMetadata(object):
    def __init__(self, version):
        self.version = version
        self.title = ""
        self.audio_signals = []
        self.audio_elements = []
        self.audio_presentations = []
        self.iat = []

    def update_version(self, version):
        self.version = version

    def update_title(self, title):
        self.title = title


class AudioPresentation(object):
    def __init__(self, name, language, id):
        self.id = id
        self.config = ""
        self.name = name
        self.language = language
        self.name_language = []
        self.elements = []

    def update_name(self, name):
        self.name = name

    def update_name_language(self, name_language):
        self.name_language = name_language

    def update_config(self, config):
        self.config = config

    def update_language(self, language):
        self.language = language

    def add_local_language_name(self, name, language):
        self.name_language.append(PresentationNameLanguage(name, language))


class PresentationNameLanguage(object):
    def __init__(self, name, language):
        self.name = name
        self.language = language


class AudioObject(object):
    def __init__(self, id):
        self.id = id
        self.name = ""
        self.classification = ""
        self.dynamic_updates = False
        self.azimuth_or_x = 0.0
        self.elevation_or_y = 0.0
        self.distance_or_z = 0.0
        self.size = 0.0
        self.size_3d = False
        self.diverge = False
        self.audio_signal = None
        self.source_gain_db = 0.0

    def update_name(self, name):
        self.name = name

    def update_classification(self, classification):
        self.classification = classification

    def update_dynamic_updates(self, dynamic_updates):
        self.dynamic_updates = dynamic_updates

    def update_coordinates(self, azimuth_or_x, elevation_or_y, distance_or_z):
        self.azimuth_or_x = azimuth_or_x
        self.elevation_or_y = elevation_or_y
        self.distance_or_z = distance_or_z

    def update_size(self, size):
        self.size = size

    def update_size_3d(self, size_3d):
        self.size_3d = size_3d

    def update_diverge(self, diverge):
        self.diverge = diverge

    def update_audio_signal(self, audio_signal_ref):
        self.audio_signal = audio_signal_ref

    def update_source_gain_db(self, source_gain_db):
        self.source_gain_db = source_gain_db


class AudioSignal(object):
    def __init__(self, id, name):
        self.id = id
        self.name = name

    def update_name(self, name):
        self.name = name


class OutputTarget(object):
    def __init__(self, target):
        self.target = target
        self.audio_signals = []


class AudioBed(object):
    def __init__(self, id, name, speaker_config):
        self.id = id
        self.name = name
        self.speaker_config = speaker_config
        self.output_targets = []

    def update_name(self, name):
        self.name = name

    def update_speaker_config(self, speaker_config):
        self.speaker_config = speaker_config


class IaT(object):
    def __init__(self, content_uuid, time_stamp):
        self.content_uuid = content_uuid
        self.time_stamp = time_stamp

# ************************************************************************************************************************************************************ #
# ************************************************************************************************************************************************************ #



"""
 AM Viewer
 Copyright (c) 2019, Dolby Laboratories Inc.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted
 provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
    and the following disclaimer in the documentation and/or other materials provided with the distribution.
 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or
    promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from pmd.pmd_const import PMD_XML_ROOT_SE_PMD

try:
    from ttk import Treeview, Scrollbar, Frame
    from Tkconstants import HORIZONTAL, VERTICAL, N,S,E,W, END
except ImportError:
    from tkinter.ttk import Treeview, Scrollbar, Frame
    from tkinter.constants import HORIZONTAL, VERTICAL, N,S,E,W, END
    
import xml.etree.ElementTree as ET

from operator import attrgetter


# The following two function should be contributed back to the pmd and adm modules
# For now they are being used here
# these functions allow either a PMD or SADM xml file to be analyzed from the command line

def isFilePmd(fileHandle):
    tree = ET.ElementTree(file=fileHandle)
    tree.getroot()
    root = tree.getroot()
    pmd_root = root.find(PMD_XML_ROOT_SE_PMD)
    fileHandle.seek(0) # rewind file so that it can still be parsed
    if pmd_root == None:
        return(False)
    else:
        return(True)

def isFileSADM(fileHandle):
    tree = ET.ElementTree(file=fileHandle)
    tree.getroot()
    root = tree.getroot()
    sadm_format_root = root.find('coreMetadata')
    fileHandle.seek(0) # rewind file so that it can still be parsed
    if sadm_format_root == None:
        return(False)
    else:
        return(True)





# Parameters
# 1. Interface
# 2. Multicast IP address

def autoscroll(sbar, first, last):
    """Hide and show scrollbar as needed."""
    first, last = float(first), float(last)
    if first <= 0 and last >= 1:
        sbar.grid_remove()
    else:
        sbar.grid()
    sbar.set(first, last)

class XML_Viewer(Frame):

    def __init__(self, master, xml=None, heading_text=None, heading_anchor=None, padding=None, cursor=None, takefocus=None, style=None, width=1000, height=1000):
        Frame.__init__(self, master, class_="XML_Viewer", width=width, height=height)
        Frame.grid_propagate(self,False)
        self._vsb = Scrollbar(self, orient=VERTICAL)
        self._hsb = Scrollbar(self, orient=HORIZONTAL)

        kwargs = {}
        kwargs["yscrollcommand"] = lambda f, l: autoscroll(self._vsb, f, l)
        kwargs["xscrollcommand"] = lambda f, l: autoscroll(self._hsb, f, l)
       
        if style is not None:
            kwargs["style"] = style
            
        if padding is not None:
            kwargs["padding"] = padding
            
        if cursor is not None:
            kwargs["cursor"] = cursor
            
        if takefocus is not None:
            kwargs["takefocus"] = takefocus

        self._treeview = Treeview(self, **kwargs)
        
        if heading_text is not None:
            if heading_anchor is not None:
                self._treeview.heading("#0", text=heading_text, anchor=heading_anchor)
            else:
                self._treeview.heading("#0", text=heading_text)

        self._treeview.bind("<<TreeviewOpen>>", self._on_open)
        self._treeview.bind("<<TreeviewClose>>", self._on_close)
        
        # Without this line, horizontal scrolling doesn't work properly.
        self._treeview.column("#0", stretch= True, minwidth=100)

        self._vsb['command'] = self._treeview.yview
        self._hsb['command'] = self._treeview.xview

        self._treeview.grid(column=0, row=0, sticky=N+S+W+E)
        self._vsb.grid(column=1, row=0, sticky=N+S)
        self._hsb.grid(column=0, row=1, sticky=E+W)
        
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self._element_tree = None
        self._item_ID_to_element = {}

        if xml is not None:
            self.parse_xml(xml)

    def _on_open(self, event):
        item_ID = self._treeview.focus()
        if item_ID not in self._item_ID_to_element: return

        node = self._item_ID_to_element[item_ID]

        self._treeview.item(item_ID, text = self._repr_of_openning_tag(node))
        
    def _on_close(self, event):
        item_ID = self._treeview.focus()
        if item_ID not in self._item_ID_to_element: return

        node = self._item_ID_to_element[item_ID]
        
        text = self._repr_of_openning_tag(node) + self._repr_of_closing_tag(node)
        self._treeview.item(item_ID, text = text)

    def parse_xml(self, xml):
        self._element_tree = ET.ElementTree(ET.fromstring(xml))
        
        self.clear()
        self._walk_xml(self._element_tree.getroot())
        
    @property
    def element_tree(self):
        return self._element_tree
    
    @element_tree.setter
    def element_tree(self, element_tree):
        self._element_tree = element_tree
        
        self.clear()
        self._walk_xml(element_tree.getroot())   
        
    def clear(self):
        self._item_ID_to_element = {}
        self._treeview.delete(*self._treeview.get_children())
        
    def _repr_of_openning_tag(self, node):
        text = "<" + node.tag

        attrs = node.attrib
        
        # list function is here necessary to provide support to Python 3
        a_names = list(attrs.keys())
        a_names.sort()

        for a_name in a_names:
            text += ' %s="' % a_name
            text += attrs[a_name]
            text += '"'

        text += ">"
        return text
        
    def _repr_of_closing_tag(self, node):
        return "</%s>"%node.tag

    def _walk_xml(self, node, depth=0, parent=""):
        text = self._repr_of_openning_tag(node) + self._repr_of_closing_tag(node)

        item = self._treeview.insert(parent, END, text = text)
        self._item_ID_to_element[item] = node
        
        if node.text:
            text = node.text.strip()
            if text != "":
                for line in text.splitlines():
                    self._treeview.insert(item, END, text = line)

        child_nodes = sorted(list(node), key=attrgetter('tag'))
        for child_node in node:
            self._walk_xml(child_node, depth+1, parent=item)

        
        if node.tail:
            tail = node.tail.strip()
            if tail != "":
                for line in tail.splitlines():
                    self._treeview.insert(parent, END, text = line)





"""
****************************************************************************************************************************************************************
****************************************************************************************************************************************************************

    ___    ____  __  ___            ____  __  _______     ____               _           __
   /   |  / __ \/  |/  /           / __ \/  |/  / __ \   / __ \_________    (_)__  _____/ /_
  / /| | / / / / /|_/ /  ______   / /_/ / /|_/ / / / /  / /_/ / ___/ __ \  / / _ \/ ___/ __/
 / ___ |/ /_/ / /  / /  /_____/  / ____/ /  / / /_/ /  / ____/ /  / /_/ / / /  __/ /__/ /_
/_/  |_/_____/_/  /_/           /_/   /_/  /_/_____/  /_/   /_/   \____/_/ /\___/\___/\__/
                                                                      /___/

****************************************************************************************************************************************************************
****************************************************************************************************************************************************************

Copyright (c) 2018, Dolby Laboratories Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

****************************************************************************************************************************************************************
****************************************************************************************************************************************************************
"""

# ************************************************************************************************************************************************************ #
# ************************************************************************************************************************************************************ #

NON_DIALOGUE_CONTENT_KIND = ["undefined", "music", "effect"]
DIALOGUE_CONTENT_KIND = ["undefined", "dialogue", "voiceover", "spoken subtitle", "audio description", "commentary", "emergency"]
MIXED_CONTENT_KIND = ["undefined", "complete main", "mixed", "hearing impaired"]
DIALOGUE_TEXT = ["nonDialogueContentKind", "dialogueContentKind", "mixedContentKind"]
COORDINATE_MODE = ["Polar", "Cartesian"]
TYPE_DEFINITION = ["Invalid", "DirectSpeakers", "Matrix", "Objects", "HOA", "Binaural"]
TYPE_LABEL = ["Invalid", "0001", "0002", "0003", "0004", "0005"]

AUDIO_PACK_STR = "AP_000"
AUDIO_PROGRAMME_STR = "APR_"
AUDIO_CONTENT_STR = "AC_"
AUDIO_OBJECT_STR = "AO_"
AUDIO_BLOCK_STR = "AB_000"
AUDIO_CHANNEL_STR = "AC_000"
AUDIO_TRACK_UID_STR = "ATU_0000"

NON_DIALOGUE_CONTENT = 0
DIALOGUE_CONTENT = 1
MIXED_CONTENT = 2

UNDEFINED = 0
MUSIC = 1
EFFECT = 2

DIALOGUE = 1
VOICEOVER = 2
SPOKEN_SUBTITLE = 3
AUDIO_DESCRIPTION = 4
COMMENTARY = 5
EMERGENCY = 6


ADM_NON_DIALOGUE_CONTENT_KIND = 0
ADM_DIALOGUE_CONTENT_KIND = 1
ADM_MIXED_CONTENT_KIND = 2

ADM_DIALOGUE_CONTENT_KIND_UNDEFINED = 0
ADM_DIALOGUE_CONTENT_KIND_DIALOGUE = 1
ADM_DIALOGUE_CONTENT_KIND_VOICEOVER = 2
ADM_DIALOGUE_CONTENT_KIND_SPOKEN_SUBTITLE = 3
ADM_DIALOGUE_CONTENT_KIND_AUDIO_DESCRIPTION = 4
ADM_DIALOGUE_CONTENT_KIND_COMMENTARY = 5
ADM_DIALOGUE_CONTENT_KIND_EMERGENCY = 6

COMPLETE_MAIN = 1
MIXED = 2
HEARING_IMPAIRED = 3

POLAR = 0
CARTESIAN = 1

DIRECTSPEAKERS = 1
MATRIX = 2
OBJECTS = 3
HOA = 4
BINAURAL = 5

MIN_AUDIO_CHANNEL = 1
MAX_AUDIO_CHANNEL = 256

DEFAULT_AUDIO_SAMPLE_RATE = 48000
DEFAULT_AUDIO_BIT_DEPTH = 24

MIN_MAX_AZIMUTH = 180
MIN_MAX_ELEVATION = 90
MIN_MAX_DISTANCE = 1

MIN_MAX_XPOS = 1
MIN_MAX_YPOS = 1
MIN_MAX_ZPOS = 1

LOUDSPEAKER_CONFIG_BS_2051_SOUND_SYSTEM_RANGE_START = 0x0
LOUDSPEAKER_CONFIG_BS_2051_SOUND_SYSTEM_A = 0x02
LOUDSPEAKER_CONFIG_BS_2051_SOUND_SYSTEM_B = 0x05
LOUDSPEAKER_CONFIG_BS_2051_SOUND_SYSTEM_D = 0x45
LOUDSPEAKER_CONFIG_BS_2051_SOUND_SYSTEM_J = 0x47
LOUDSPEAKER_CONFIG_BS_2051_SOUND_SYSTEM_RANGE_END = 0x100

LOUDSPEAKER_CONFIG_COMMON_USE_RANGE_START = 0x100
LOUDSPEAKER_CONFIG_COMMON_USE_2_0 = 0x102
LOUDSPEAKER_CONFIG_COMMON_USE_3_0 = 0x103
LOUDSPEAKER_CONFIG_COMMON_USE_5_0 = 0x105
LOUDSPEAKER_CONFIG_COMMON_USE_5_1 = 0x106
LOUDSPEAKER_CONFIG_COMMON_USE_5_0_4 = 0x145
LOUDSPEAKER_CONFIG_COMMON_USE_5_1_4 = 0x146
LOUDSPEAKER_CONFIG_COMMON_USE_7_0_4 = 0x147
LOUDSPEAKER_CONFIG_COMMON_USE_7_1_4 = 0x148
LOUDSPEAKER_CONFIG_COMMON_USE_9_0_6 = 0x169
LOUDSPEAKER_CONFIG_COMMON_USE_9_1_6 = 0x16a
LOUDSPEAKER_CONFIG_COMMON_USE_RANGE_END = 0x200

LOUDSPEAKER_CONFIG_HORIZONTAL_CHANNEL_MASK = 0xf
LOUDSPEAKER_CONFIG_VERTICAL_CHANNEL_MASK = 0xf0

LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_LABELS = ["M+030", "M-030", "M+000", "M+090", "M-090", "M+135", "M-135"]
LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_ALT_LABELS = ["M+030", "M-030", "M+000", "M+110", "M-110"]
LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_AZIMUTH = ["30.0", "-30.0", "0.0", "90.0", "-90.0", "135.0", "-135.0"]
LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_ALT_AZIMUTH = ["30.0", "-30.0", "0.0", "110.0", "-110.0"]
LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_ELEVATION = ["0.0", "0.0", "0.0", "0.0", "0.0", "0.0", "0.0"]
LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_DISTANCE = ["1.0", "1.0", "1.0", "1.0", "1.0", "1.0", "1.0"]

LOUDSPEAKER_CONFIG_BS_2051_HORIZONTAL_SPEAKER_ALT_SWITCH = 5

LOUDSPEAKER_CONFIG_BS_2051_VERTICAL_SPEAKER_LABELS = ["U+045", "U-045", "U+135", "U-135"]
LOUDSPEAKER_CONFIG_BS_2051_VERTICAL_SPEAKER_ALT_LABELS = ["U+030", "U-030", "U+110", "U-110"]
LOUDSPEAKER_CONFIG_BS_2051_VERTICAL_SPEAKER_AZIMUTH = ["45.0", "-45.0", "135.0", "-135.0"]
LOUDSPEAKER_CONFIG_BS_2051_VERTICAL_SPEAKER_ALT_AZIMUTH = ["30.0", "-30.0", "110.0", "-110.0"]
LOUDSPEAKER_CONFIG_BS_2051_VERTICAL_SPEAKER_ELEVATION = ["30.0", "30.0", "30.0", "30.0"]
LOUDSPEAKER_CONFIG_BS_2051_VERTICAL_SPEAKER_DISTANCE = ["1.0", "1.0", "1.0", "1.0"]

LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_LABELS = ["RC_L", "RC_R", "RC_C", "RC_LFE", "RC_Ls", "RC_Rs", "RC_Lrs", "RC_Rrs"]
LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_XPOS = ["-1.0", "1.0", "0.0", "-1.0", "-1.0", "1.0", "-1.0", "1.0"]
LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_YPOS = ["1.0", "1.0", "1.0", "1.0", "0.0", "0.0", "-1.0", "1.0"]
LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_ALT_YPOS = ["1.0", "1.0", "1.0", "1.0", "-1.0", "1.0"]
LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_ZPOS = ["0.0", "0.0", "0.0", "0.0", "0.0", "0.0", "0.0", "0.0"]

LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_LFE_INDEX_VALUE = 3
LOUDSPEAKER_CONFIG_COMMON_USE_HORIZONTAL_SPEAKER_ALT_SWITCH = 6

LOUDSPEAKER_CONFIG_COMMON_USE_VERTICAL_SPEAKER_LABELS = ["RC_Ltf", "RC_Rtf", "RC_Ltr", "RC_Rtr"]
LOUDSPEAKER_CONFIG_COMMON_USE_VERTICAL_SPEAKER_XPOS = ["-1.0", "1.0", "-1.0", "1.0"]
LOUDSPEAKER_CONFIG_COMMON_USE_VERTICAL_SPEAKER_YPOS = ["1.0", "1.0", "-1.0", "-1.0"]
LOUDSPEAKER_CONFIG_COMMON_USE_VERTICAL_SPEAKER_ZPOS = ["1.0", "1.0", "1.0", "1.0"]

ADM_LOG_FILE = 'adm_tool.log'
LIMITS_FILE = 'limits.txt'
MSG_VALIDATION_PASS = 'Validation stage passed'
MSG_VALIDATION_FAIL = 'Validation stage failed'
MSG_VALIDATION_ABRT = ' *** Aborting ***'
MSG_INVALID_STUB = 'Ended up in code that should never execute in '
MSG_TYP_LBL_IS_DS = ' is type DirectSpeakers'
MSG_TYP_LBL_IS_OB = ' is type Object'
MSG_FND_EL = 'Found '
MSG_MIN_EL = ', minimum entries allowed is '
MSG_MAX_EL = ', maximum entries allowed is '
MSG_FND_NO = ' could not find '
MSG_LIM_NO = ' info in limits file *** Aborting ***'
MSG_CONTENT_INFO = 'Content kind'

ADM_XML_MODE_FILE = 0
ADM_XML_MODE_STRING = 1

ADM_XML_TYP_DS = '_0001'
ADM_XML_TYP_OB = '_0003'
ADM_XML_INT_TYP_DS = 1
ADM_XML_INT_TYP_OB = 3

ADM_XML_APR_ELN = 'audioProgramme'
ADM_XML_APR_ELN_AT_NM = 'audioProgrammeName'
ADM_XML_APR_ELN_AT_LN = 'audioProgrammeLanguage'
ADM_XML_APR_ELN_AT_ID = 'audioProgrammeID'
ADM_XML_APR_ELN_SE_PL = 'audioProgrammeLabel'
ADM_XML_APR_ELN_SE_PL_AT_LN = 'language'
ADM_XML_APR_ELN_SE_CR = 'audioContentIDRef'
ADM_XML_APR_ELN_SE_LM = 'loudnessMetadata'

ADM_XML_ACO_ELN = 'audioContent'
ADM_XML_ACO_ELN_AT_ID = 'audioContentID'
ADM_XML_ACO_ELN_AT_NM = 'audioContentName'
ADM_XML_APR_ELN_SE_DG = 'dialogue'
ADM_XML_APR_ELN_SE_DG_AT_ND = 'nonDialogueContentKind'
ADM_XML_APR_ELN_SE_DG_AT_DC = 'dialogueContentKind'
ADM_XML_APR_ELN_SE_DG_AT_MC = 'mixedContentKind'

ADM_XML_APR_ELN_SE_AO = 'audioObjectIDRef'

ADM_XML_AOB_ELN = 'audioObject'
ADM_XML_AOB_ELN_AT_ID = 'audioObjectID'
ADM_XML_AOB_ELN_AT_NM = 'audioObjectName'
ADM_XML_AOB_ELN_SE_AP = 'audioPackFormatIDRef'
ADM_XML_AOB_ELN_SE_AT = 'audioTrackUIDRef'
ADM_XML_AOB_ELN_SE_HL = 'headLocked'
ADM_XML_AOB_ELN_SE_GN = 'gain'

ADM_XML_APF_ELN = 'audioPackFormat'
ADM_XML_APF_ELN_AT_ID = 'audioPackFormatID'
ADM_XML_APF_ELN_AT_NM = 'audioPackFormatName'
ADM_XML_APF_ELN_AT_TL = 'typeLabel'
ADM_XML_APF_ELN_SE_AC = 'audioChannelFormatIDRef'

ADM_XML_ACF_ELN = 'audioChannelFormat'
ADM_XML_ACF_ELN_AT_ID = 'audioChannelFormatID'
ADM_XML_ACF_ELN_AT_NM = 'audioChannelFormatName'
ADM_XML_ACF_ELN_SE_AB = 'audioBlockFormat'
ADM_XML_ACF_ELN_SE_AB_AT_ID = 'audioBlockFormatID'
ADM_XML_ACF_ELN_SE_AB_SE_CT = 'cartesian'
ADM_XML_ACF_ELN_SE_AB_SE_SL = 'speakerLabel'
ADM_XML_ACF_ELN_SE_AB_SE_PS = 'position'
ADM_XML_ACF_ELN_SE_AB_SE_PS_AT_CO = 'coordinate'
ADM_XML_ACF_ELN_SE_AB_SE_PS_AT_XC = 'X'
ADM_XML_ACF_ELN_SE_AB_SE_PS_AT_YC = 'Y'
ADM_XML_ACF_ELN_SE_AB_SE_PS_AT_ZC = 'Z'
ADM_XML_ACF_ELN_SE_AB_SE_HR = 'headphoneRender'
ADM_XML_ACF_ELN_SE_AB_SE_HR_AT_BP = 'Bypass'
ADM_XML_ACF_ELN_SE_AB_SE_HR_AT_MD = 'Mode'

ADM_XML_ATU_ELN = 'audioTrackUID'
ADM_XML_ATU_ELN_AT_ID = 'UID'
ADM_XML_ATU_ELN_SE_CF = 'audioChannelFormatIDRef'
ADM_XML_ATU_ELN_SE_PF = 'audioPackFormatIDRef'

ADM_XML_INDENT = "  "
PMD_XML_MODE_FILE = 0
PMD_XML_MODE_STRING = 1

# ************************************************************************************************************************************************************ #
# ************************************************************************************************************************************************************ #



#!/usr/bin/env python3

"""
 AM Viewer
 Copyright (c) 2019, Dolby Laboratories Inc.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted
 provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
    and the following disclaimer in the documentation and/or other materials provided with the distribution.
 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or
    promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from tkinter import *
import time
import random
import threading
import platform
import os
import signal
import sys
import subprocess
import binascii
import zlib
import netifaces
import scapy.all as scapy
import argparse
import queue
import copy
import wave
import errno
from pmd.pmd_tool import parse_pmd_xml
from pmd.pmd_tool import populate_model_from_xml
from pmd.pmd_tool import populate_model_from_adm
from pmd.pmd_tool import PMD_XML_MODE_FILE
from pmd.pmd_tool import ADM_XML_MODE_FILE
from adm.adm_tool import ADM_XML_MODE_STRING
from pmd.pmd_classes import AudioObject
from pmd.pmd_classes import AudioBed
from pmd.pmd_tool import check_parameter_type
from pmd.pmd_const import PRESENTATION_CONFIG_TEXT
from pmd.pmd_const import PRESENTATION_CONFIG_EQUIV
from pmd.pmd_const import LOUDSPEAKER_CONFIG_COMMON_USE_TEXT_NAMES
from pmd.pmd_const import LOUDSPEAKER_CONFIG_2_0
from pmd.pmd_const import LOUDSPEAKER_CONFIG_5_1
from pmd.pmd_const import OBJECT_CLASSES
from am_viewer.am_xml_viewer import XML_Viewer
from am_viewer.am_xml_viewer import isFilePmd
from am_viewer.am_xml_viewer import isFileSADM
import aoip_services.aoip_discovery
import aoip_services.multicast

__version__ = "3.1"

class AudioObjectHeadings:
    TYPE = 0
    NAME = 1
    DIVERGE = 2
    CH = 3
    GAIN = 4
    X = 5
    Y = 6
    Z = 7

class PresentationHeadings:
    NAME = 0
    NAME_LANG = 1
    BED_ELEMENT = 2
    TARGET_CFG = 3
    PRES_LANG = 4

class AudioBedHeadings:
    SOURCE = 0
    CONFIG = 1
    NAME = 2
    START = 3
    END = 4


# Helper functions

def get_cmd_stdio(args):
    import subprocess
    run_output = subprocess.run(args, stdout=subprocess.PIPE)
    text = str(run_output.stdout.decode("utf-8"))
    text = text.replace('\\n', '\n')
    text = text.replace('\\r', '\r')
    text = text.replace('\\t', '\t')
    return(text)


def gain_from_db(db):
    return 10 ** (db / 20)

def get_word(payload, index, subframe_mode):
    if subframe_mode:
        return(int.from_bytes(payload[index*2:(index*2)+3], byteorder='big') >> 4)
    else:
        return(int.from_bytes(payload[index:index+3], byteorder='big') >> 4)


def endian_swap24(array24):
    swapped24 = bytearray(b'')
    for i in range (0, len(array24), 3):
        swapped24.append(array24[i + 2])
        swapped24.append(array24[i + 1])
        swapped24.append(array24[i])
    return swapped24

def check_xml_complete():
    with open('pmd.xml') as f:
        lines = f.readlines()
        if len(lines) < 25:
            return False
        for line in lines[::-1]:
            if line.find("<IAT>") > 0:
                return True
        return False


def pack_20bits(array24):
    output_nibble_hi_lo = 0
    output_nibble_count = 0
    input_word_count = len(array24)
    array20 = bytearray(b'')

    for i in range(0, input_word_count, 3):
        input_nibbles = []
        for j in range(0,3):
            input_nibbles.append((array24[i+j] >> 4) & 0x0f)
            input_nibbles.append(array24[i+j] & 0x0f)
        # Write out 5 nibbles
        for nibble in input_nibbles:
            if (output_nibble_count != 5):
                if output_nibble_hi_lo == 1:
                    array20[-1] = array20[-1] | nibble
                else:
                    array20.append(nibble << 4)
                output_nibble_hi_lo = 1 - output_nibble_hi_lo
                output_nibble_count = output_nibble_count + 1
            else:
                output_nibble_count = 0

    return(array20)

def subframe_to_framemode(array24):
    if (len(array24) % 6) != 0:
        print("Warning: trying to convert from subframe mode with incomplete # words")
        limit = (len(array24) // 6) * 6 #uses floor division //
    else:
        limit = len(array24)
    frame_array = bytearray(b'')
    for i in range(0, limit, 6):
        for j in range(i, i+3):
            frame_array.append(array24[j])
    return(bytes(frame_array))


def clean_up_temp_files(filelist):
    for tmpfile in filelist:
        try:
            os.remove(tmpfile)
        except:
            pass



# Debug helpers -- This can be deleted after development

# --- - chunking helpers
def chunks(seq, size):
    '''Generator that cuts sequence (bytes, memoryview, etc.)
     into chunks of given size. If `seq` length is not multiply
     of `size`, the lengh of the last chunk returned will be
     less than requested.

     >>> list( chunks([1,2,3,4,5,6,7], 3) )
     [[1, 2, 3], [4, 5, 6], [7]]
    '''
    d, m = divmod(len(seq), size)
    for i in range(d):
        yield seq[i*size:(i+1)*size]
    if m:
        yield seq[d*size:]

def chunkread(f, size):
    '''Generator that reads from file like object. May return less
     data than requested on the last read.'''
    c = f.read(size)
    while len(c):
        yield c
        c = f.read(size)

def genchunks(mixed, size):
    '''Generator to chunk binary sequences or file like objects.
     The size of the last chunk returned may be less than
     requested.'''
    if hasattr(mixed, 'read'):
        return chunkread(mixed, size)
    else:
        return chunks(mixed, size)
# --- - /chunking helpers

def dump(binary, size=2, sep=' '):
    '''
    Convert binary data (bytes in Python 3 and str in
    Python 2) to hex string like '00 DE AD BE EF'.
    `size` argument specifies length of text chunks
    and `sep` sets chunk separator.
    '''
    hexstr = binascii.hexlify(binary)
    if sys.version_info[0] >= 3:
        hexstr = hexstr.decode('ascii')
    return sep.join(chunks(hexstr.upper(), size))


def my_hexdump(data, stride = 8):
    '''
    Generator that produces strings:

    '00000000: 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00  ................'
    '''
    generator = genchunks(data, stride)
    for addr, d in enumerate(generator):
        # 00000000:
        line = '%04X ' % (addr*stride)
        # 00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00
        dumpstr = dump(d)
        line += dumpstr[:3 * (int(stride /2))]
        if len(d) > 3:  # insert separator if needed
            line += ' ' + dumpstr[3 * (int(stride / 2)):]
        # ................
        # calculate indentation, which may be different for the last line
        pad = 2
        #if len(d) < 16:
        #  pad += 3*(16 - len(d))
        #if len(d) <= 8:
        #  pad += 1
        line += ' '*pad

        for byte in d:
        # printable ASCII range 0x20 to 0x7E
            if sys.version_info[0] < 3:
                byte = ord(byte)
            if 0x20 <= byte <= 0x7E:
                line += chr(byte)
            else:
                line += '.'
        print(line)


# Main Classes

## Payload Class

class StoppableThread(threading.Thread):

    def __init__(self, **kwargs):
        super(StoppableThread, self).__init__(**kwargs)
        self._stop_event = threading.Event()
        self._reset_event = threading.Event()
        self._lock = threading.Lock()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def reset(self):
        self._reset_event.set()

    def is_reset(self):
        return self._reset_event.is_set()

    def clear_reset(self):
        self._reset_event.clear()

    def get_lock(self):
        self._lock.acquire()

    def release_lock(self):
        self._lock.release()

    def lock(self):
        return(self._lock)

#def hasSync(packet):
#	pa_bytes = bytes([0x6F,0x87,0x20])
#	return(bytes(packet["RTP"].payload).find(pa_bytes))


def get_next_pmd_timeslot(sampleCount):
    # Maintain 4 samples of zeros between burst and find start of the next time-slot
    return ((((sampleCount + 4) // 160) + 1) * 160) - sampleCount



def get_pmd_xml_from_wav(frame):
    pmd_tool_exec = os.path.dirname(__file__) + "/pmd_tool." + platform.system()
    if (platform.system() == 'Windows'):
        pmd_tool_exec = pmd_tool_exec + '.exe'
    if not os.path.exists(pmd_tool_exec):
        print("__file__:",__file__,"  pmd_tool path:", pmd_tool_exec)
        print("Platform", platform.system(), "not supported", file=sys.stderr)
        exit(-4)
    # Add preamble


    with wave.open("pmd.wav", "wb") as wave_file:
        wave_file.setnchannels(2)
        wave_file.setsampwidth(3)  # 24 bits / 3 bytes
        wave_file.setframerate(48000)

        # Start with zero padding from vSync
        sampleCount = 32
        wave_file.writeframes(bytes(sampleCount * 6))
        # Now add payloads
        for payload in frame.payloads:
            # swap to little endian as payload 24 is big endian as required for KLV
            # unfortunately wav is little so we have to swap here
            payload24_le = endian_swap24(payload)
            wave_file.writeframes(payload24_le)
            sampleCount = sampleCount + (len(payload24_le) // 6)
            # Add post frame seperator
            zeroPadSamples = get_next_pmd_timeslot(sampleCount)
            wave_file.writeframes(bytes(zeroPadSamples * 6))
            sampleCount = sampleCount + zeroPadSamples

    args = [pmd_tool_exec, "-i", "pmd.wav", "-o", "pmd.xml"]
    subprocess.run(args)
    if not os.path.exists("pmd.xml"):
        print("KLV to XML conversion failed. PMD is probably corrupted", file=sys.stderr)
        return False
    return True


def get_pmd_xml_from_klv(frame, pack20_flag):
    pmd_tool_exec = os.path.dirname(__file__) + "/pmd_tool." + platform.system()
    if platform.system() == 'Windows':
        pmd_tool_exec = pmd_tool_exec + '.exe'
    if not os.path.exists(pmd_tool_exec):
        print("__file__:",__file__,"  pmd_tool path:", pmd_tool_exec)
        print("Platform", platform.system(), "not supported", file=sys.stderr)
        exit(-4)

    if pack20_flag:
        klv_payload = pack_20bits(frame.payload)
    else:
        klv_payload = frame.payloads[0]
    with open("pmd.klv", "wb") as klv_file:
        klv_file.write(klv_payload)
    try:
        os.remove("pmd.xml")
    except:
        pass
    args = [pmd_tool_exec, "-i", "pmd.klv", "-o", "pmd.xml"]
    subprocess.run(args)
    if not os.path.exists("pmd.xml"):
        print("KLV to XML conversion failed. PMD is probably corrupted", file=sys.stderr)
        return False
    return True


def get_sadm_xml(frame):
    index = 0
    if frame.sADM_assemble_flag == 1:
        assemble_info = get_word(frame.payloads[0], index, False)
        in_timeline_flag = (assemble_info >> 4) & 0x3
        # Only support full frame mode so this must be 0
        if (in_timeline_flag != 0):
            return(None)
        #Commenting the next two elements out for speed as they are ignored
        #track_numbers = (assemble_info >> 6) & 0x3f
        #track_ID = (assemble_info >> 12) & 0x3f
        index = index + 3
    if frame.sADM_format_flag == 1:
        format_info = get_word(frame.payloads[0], index, False)
        format_type = format_info >> 4 & 0xf
        index = index + 3
    else:
        format_type = 0
    if format_type == 1:
        gzip_data = pack_20bits(frame.payloads[0][index:])
        try:
            adm_xml = zlib.decompress(gzip_data, 15 + 32)
        except:
            adm_xml = None
    else:
        adm_xml = frame.payloads[0][index:]
    return adm_xml

class State:
    IDLE = 16
    GOT_LEFT_BLANKING = 32
    GOT_RIGHT_BLANKING = 64
    GOT_FRAME_MODE_BLANKING = 128
    GOT_PA = 1
    GOT_SYNC = 2
    GOT_HEADER = 3
    GOT_PMD_FRAME = 4
    GOT_END_BLANKING = 5

class pmdDeframer:
 
    subframe_mode = None
    got_header = False
    length20=0
    length24=0
    data_type=0

    payloads = b''
    left_zero_count = 0
    right_zero_count = 0
    frame_mode_zero_count = 0
    end_zero_count = 0
    state = State.IDLE
    last_sequnce_no = None
    new_frames = []
    active_channel = False

    pa = 0x6f872
    pb = 0x54e1f
    zero_threshold = 350
    intraframe_zero_threshold = 160

    class NewFrame:
        payload = None
        format = None
        subframe_mode = None
        right_not_left = None
        sADM_assemble_flag = None
        sADM_format_flag = None

        def __init__(self):
            self.payloads = []
            self.format = None
            self.subframe_mode = None
            self.right_not_left = None
            self.sADM_assemble_flag = None
            self.sADM_format_flag = None

        def is_sADM(self):
            if self.format == "SADM":
                return True
            else:
                return False

        def is_pmd(self):
            if self.format == "PMD":
                return True
            else:
                return False

        def is_AESX242(self):
            if self.format == "AESX242":
                return True
            else:
                return False

        def get_sADM_assemble_flag(self):
            return self.sADM_assemble_flag

        def get_sADM_format_flag(self):
            return self.sADM_format_flag


    def __init__(self):
        self.subframe_mode = None
        self.got_header = False
        self.length20=0
        self.length24=0
        self.data_type=0

        self.payloads = b''
        self.left_zero_count = 0
        self.right_zero_count = 0
        self.frame_mode_zero_count = 0
        self.end_zero_count = 0
        self.state = State.IDLE
        self.last_sequnce_no = None
        self.new_frames = []
        self.next_frame = None
        self.active_channel = False

        self.pa = 0x6f872
        self.pb = 0x54e1f
        self.zero_threshold = 350
        self.intraframe_zero_threshold = 160



    def reset(self):
        self.subframe_mode = None
        self.got_header = False
        self.length20 = 0
        self.length24 = 0
        self.data_type = 0
        self.payloads = b''
        self.left_zero_count = 0
        self.right_zero_count = 0
        self.frame_mode_zero_count = 0
        self.end_zero_count = 0
        self.state = State.IDLE
        self.last_sequnce_no = None
        self.active_channel = False
        self.next_frame = None

    def receivePacket(self, packet, codec):
        if codec == "AM824":
            self.receiveSMPTEPacket(packet)
        elif codec == "smpte336m":
            self.receiveAESX242Packet(packet)
        else:
            raise("Unknown format")

    def receiveAESX242Packet(self, packet):
        packet["UDP"].payload = scapy.RTP(packet["Raw"].load)
        # detect marker bit to see if this is the first packet of the KLVunit
        if packet["RTP"].marker:
            # See if we already have a payload, if so then we can process it
            if len(self.payloads) > 0:
                next_frame = self.NewFrame()
                next_frame.format = "AESX242"
                next_frame.payloads = [self.payloads]
                self.new_frames.append(next_frame)
            self.payloads = bytes(packet["RTP"].payload)
        else:
            self.payloads = self.payloads + bytes(packet["RTP"].payload)

    # Receive raw packets as they are received
    def receiveSMPTEPacket(self, packet):

        # Force packet to be interpreted as RTP
        packet["UDP"].payload = scapy.RTP(packet["Raw"].load)
        packet_payload = bytes(packet["RTP"].payload)
        right_not_left = True
        if self.last_sequnce_no is None:
            self.last_sequnce_no = packet["RTP"].sequence
        elif packet["RTP"].sequence == (self.last_sequnce_no + 1) & 0xffff:
            self.last_sequnce_no = packet["RTP"].sequence
        else:
            self.reset()
        for index in range(0, len(packet_payload), 4):
            # If subframe mode is detected
            right_not_left = not right_not_left
            if self.subframe_mode:
                if not self.active_channel:
                    self.active_channel = True
                    continue
                else:
                    self.active_channel = False

            # Because we are aligned to the actual sync word we drop the fourth/last byte
            next_word = packet_payload[index + 1:index + 4]
            next_sample = int.from_bytes(next_word, byteorder='big') >> 4

            # Search for blanking period: 350 zeros
            if self.state & State.IDLE:
                if next_sample == 0:
                    if right_not_left:
                        self.right_zero_count = self.right_zero_count + 1
                        if self.right_zero_count >= self.zero_threshold:
                            self.state = self.state | State.GOT_RIGHT_BLANKING
                    else:
                        self.left_zero_count = self.left_zero_count + 1
                        if self.left_zero_count >= self.zero_threshold:
                            self.state = self.state | State.GOT_LEFT_BLANKING
                    self.frame_mode_zero_count = self.frame_mode_zero_count + 1

                    # If we've got enough zeros to know the first burst is next then move to next state
                    if self.frame_mode_zero_count >= self.zero_threshold:
                        self.state = self.state | State.GOT_FRAME_MODE_BLANKING

                    # Check to see if we have a PMD frame to release because we know there are no new payloads
                    if self.next_frame is not None and self.next_frame.is_pmd():
                        if self.next_frame.subframe_mode:
                            if self.next_frame.right_not_left and (self.right_zero_count > self.intraframe_zero_threshold):
                                self.next_frame.payloads.append(self.payloads)
                                self.new_frames.append(self.next_frame)
                                self.next_frame = None
                            if (not self.next_frame.right_not_left) and (self.left_zero_count > self.intraframe_zero_threshold):
                                self.next_frame.payloads.append(self.payloads)
                                self.new_frames.append(self.next_frame)
                                self.next_frame = None
                        else:
                            if self.right_zero_count > self.intraframe_zero_threshold and self.left_zero_count > self.intraframe_zero_threshold:
                                self.next_frame.payloads.append(self.payloads)
                                self.new_frames.append(self.next_frame)
                                self.next_frame = None
                else:
                    # reset parser until we find zeros
                    if right_not_left:
                        if next_sample == self.pa and (self.state & State.GOT_RIGHT_BLANKING):
                            self.state = State.GOT_PA
                            self.payloads = self.payloads + next_word
                        else:
                            self.right_zero_count = 0
                            self.frame_mode_zero_count = 0
                            self.state = self.state & ~(State.GOT_RIGHT_BLANKING + State.GOT_FRAME_MODE_BLANKING)
                    else:
                        if next_sample == self.pa and (self.state & State.GOT_LEFT_BLANKING):
                            self.state = State.GOT_PA
                            self.payloads = self.payloads + next_word
                        else:
                            self.left_zero_count = 0
                            self.frame_mode_zero_count = 0
                            # Remove word as not used
                            self.state = self.state & ~(State.GOT_LEFT_BLANKING + State.GOT_FRAME_MODE_BLANKING)

            elif self.state == State.GOT_PA:
                self.payloads = self.payloads + next_word
                frame_pos = len(self.payloads)
                if frame_pos == 6:
                    if next_sample == self.pb and right_not_left:
                        self.subframe_mode = False
                        self.state = State.GOT_SYNC
                elif frame_pos == 9:
                    if next_sample == self.pb:
                        self.subframe_mode = True
                        self.state = State.GOT_SYNC
                        # remove middle word from payload
                        self.payloads = self.payloads[0:3] + self.payloads[6:9]
                    else:
                        self.reset()
                else:
                    # should be reachable as reset above should be reached first
                    self.reset()

            elif self.state == State.GOT_SYNC:
                self.payloads = self.payloads + next_word
                if self.next_frame is None:
                    self.next_frame = self.NewFrame()
                # Check sequence otherwise if we are out of order mid-frame then errors
                # are certain so reset state machine and reacquire sync
                frame_pos = len(self.payloads)

                if frame_pos == 15:
                    Pc = get_word(self.payloads, 6, False)
                    self.data_type = (Pc & 0x01f0) >> 4
                    # If extended data type then add in Pe
                    if self.data_type == 31:
                        self.data_type = self.data_type + get_word(self.payloads, 12, False) & 0xffff
                    # 8 bits, scale up to 24 bits because we are not using packed 20 bits yet
                    Pd = get_word(self.payloads, 9, False)
                    self.length20 = int(Pd / 8)
                    self.length24 = int(Pd / 20) * 3
                    # If subframe mode then we need twice as many samples to hold frame
                    if self.data_type == 27:
                        self.next_frame.format = "PMD"
                    if self.data_type == 32:
                        self.next_frame.format = "SADM"
                        self.next_frame.sADM_assemble_flag = (Pc >> 13) & 0x1
                        self.next_frame.sADM_format_flag = (Pc >> 14) & 0x1
                    # adjust for Pe & Pf
                    #if self.data_type > 30:
                        # adjust for Pe and Pf
                    #    self.length20 = self.length20 - 5
                    #    self.length24 = self.length24 - 6
                    self.state = State.GOT_HEADER
                    if (not (self.next_frame.is_pmd())) and (not (self.next_frame.is_sADM())):
                        # detected wrong SMPTE data type
                        self.reset()

            elif self.state == State.GOT_HEADER:
                self.payloads = self.payloads + next_word
                # See if we have complete frame
                if len(self.payloads) - 12 >= self.length24:
                    self.payloads = self.payloads[0: self.length24 + 12]
                    self.next_frame.subframe_mode = self.subframe_mode

                    # Don't signal new frame if pmd because we might have more payloads
                    if self.next_frame.is_pmd():
                        self.next_frame.payloads.append(self.payloads)
                        self.next_frame.right_not_left = right_not_left
                        # Hold reference to current frame across reset
                        temp_next_frame = self.next_frame
                        self.reset()
                        self.next_frame = temp_next_frame
                        # Set state according to location of
                        if self.next_frame.right_not_left:
                            self.state = self.state + State.GOT_RIGHT_BLANKING
                        else:
                            self.state = self.state + State.GOT_LEFT_BLANKING
                    else:
                        # Strip header according to data type because of Pe/Pf
                        # [0] should be first word of SMPTE 337 payload
                        if self.data_type > 30:
                            self.next_frame.payloads.append(self.payloads[18:])
                        else:
                            self.next_frame.payloads.append(self.payloads[12:])
                        self.new_frames.append(self.next_frame)
                        self.reset()
            else:
                raise RuntimeError("Unknown State")

    def getFrame(self):
        if len(self.new_frames) > 0:
            return self.new_frames.pop(0)
        else:
            return None

    def haveFrame(self):
        return len(self.new_frames) > 0

# Main GUI App Class

class PmdAdmDisplayGUI:

    numObjects = 6
    numPresentations = 6
    numBeds = 2
    frameRelief = RAISED
    frameBg = "blue"
    frameBorderwidth = 5
    running = False
    stop = False
    interface = None
    interfaceName = ""
    mainThread = None
    serviceList = []
    serviceNameList = []
    audioList = []
    audioNameList = []
    selectedService = None
    audioService = None
    pmdIndVar = None
    sadmIndVar = None
    aesx242IndVar = None
    XMLViewerRequest = False
    discoveryService = None
    start_time = 0
    numCapturePackets = 42
    debug = False
    messageQueue = None
    multicast = None
    playbackPres = None
    playbackPresVar = None
    model = None
    audioPipeLine = None
    mixMatrix = None
    haveGstreamer = False

    class Indicators:
        pmd = "gray"
        sadm = "gray"
        aesx242 = "gray"
        frame = "gray"
        subframe = "gray"
        gui = None

        def __init__(self, gui):
            self.gui = gui

        def reset(self):
            self.pmd = "gray"
            self.sadm = "gray"
            self.aesx242 = "gray"
            self.frame = "gray"
            self.subframe = "gray"
            self.gui.post("updateInd", None)

        def rxMetadata(self):
            return self.pmd =="light green" or self.sadm =="light green" or self.aesx242 =="light green"

        def pmdOn(self):
            if not self.pmd == "light green":
                self.pmd = "light green"
                self.sadm = "gray"
                self.aesx242 = "gray"
                self.gui.post("updateInd", None)

        def pmdError(self):
            if not self.pmd == "red":
                self.pmd = "red"
                self.sadm = "gray"
                self.aesx242 = "gray"
                self.gui.post("updateInd", None)

        def sadmOn(self):
            if not self.sadm == "light green":
                self.sadm = "light green"
                self.pmd = "gray"
                self.aesx242 = "gray"
                self.gui.post("updateInd", None)

        def sadmError(self):
            if not self.sadm == "red":
                self.sadm = "red"
                self.pmd = "gray"
                self.aesx242 = "gray"
                self.gui.post("updateInd", None)

        def aesx242On(self):
            if not self.aesx242 == "light green":
                self.aesx242 = "light green"
                self.pmd = "gray"
                self.sadm = "gray"
                self.frame = "gray"
                self.subframe = "gray"
                self.gui.post("updateInd", None)

        def aesx242Error(self):
            if not self.aesx242 == "red":
                self.aesx242 = "red"
                self.pmd = "gray"
                self.sadm = "gray"
                self.frame = "gray"
                self.subframe = "gray"
                self.gui.post("updateInd", None)


        def frameMode(self):
            if not self.frame == "light green":
                self.frame = "light green"
                self.subframe = "gray"
                self.gui.post("updateInd", None)

        def subFrameMode(self):
            if not self.subframe == "light green":
                self.subframe = "light green"
                self.frame = "gray"
                self.gui.post("updateInd", None)


    def __init__(self, master, xml_filename, sdp_filename, debug_mode):

        master.title("Audio Metadata Viewer v" + __version__)
        self.debug = debug_mode

        self.indFrame = Frame(master, relief=self.frameRelief, borderwidth=self.frameBorderwidth)
        self.indFrame.pack(fill=BOTH, expand=True)

        self.pmdInd = Label(self.indFrame, text='PMD', variable=self.pmdIndVar, fg='black')
        self.pmdInd.pack(side=LEFT)
        self.sadmInd = Label(self.indFrame, text='sADM', variable=self.sadmIndVar, fg='black')
        self.sadmInd.pack(side=LEFT)
        self.aesx242Ind = Label(self.indFrame, text='AES-X242', variable=self.aesx242IndVar, fg='black')
        self.aesx242Ind.pack(side=LEFT)

        self.FrameInd = Label(self.indFrame, text='Frame', fg='black')
        self.FrameInd.pack(side=LEFT)
        self.SubFrameInd = Label(self.indFrame, text='Subframe', fg='black')
        self.SubFrameInd.pack(side=LEFT)

        self.indicators = self.Indicators(self)

        audio_beds_label = Label(master, text="Audio Beds")
        audio_beds_label.pack(fill=X)

        self.abFrame = Frame(master, relief= self.frameRelief, borderwidth=self.frameBorderwidth)
        self.abFrame.pack(fill=BOTH, expand=True)

        abLabels = [ "Source", "Config", "Name", "Start", "End" ]

        for abColumn in range(1,6):
            abLabel = Label(self.abFrame, text = abLabels[abColumn - 1])
            abLabel.grid(row = 0, column = abColumn)

        self.bedFields = [[], []]
        for bed in range(0,self.numBeds):

            bed_label = Label(self.abFrame, text = str(bed+1))
            bed_label.grid(row = bed + 1, column = 0)

            for abColumn in range(1,6):
                self.bedFields[bed].append(Label(self.abFrame, text = "None", relief=SUNKEN, bg = '#B3B4C8'))
                self.bedFields[bed][-1].grid(row = bed + 1, column = abColumn, sticky=N+E+S+W)

        for abColumn in range(0,6):
            self.abFrame.grid_columnconfigure(abColumn,minsize = 100)

        audio_objects_label = Label(master, text="Audio Objects")
        audio_objects_label.pack(fill=X)

        self.aoFrame = Frame(master, relief=RAISED, borderwidth=5)
        self.aoFrame.pack(fill=BOTH, expand=True)

        aoLabels = [ "Type", "Name", "Diverge", "Ch", "Gain(dB)", "X", "Y", "Z" ]

        for aoColumn in range(1, len(aoLabels) + 1):
            aoLabel = Label(self.aoFrame, text = aoLabels[aoColumn - 1])
            aoLabel.grid(row = 0, column = aoColumn)

        self.objectFields = [[] for i in range(self.numObjects)]
        for aoObject in range(0,self.numObjects):
            audio_object_number_label = Label(self.aoFrame, text = str(aoObject + 1))
            audio_object_number_label.grid(row = aoObject + 1, column = 0)

            for aoColumn in range(1, len(aoLabels) + 1):
                self.objectFields[aoObject].append(Label(self.aoFrame, justify = LEFT, text = "None", relief=SUNKEN, bg = '#B3B4C8'))
                self.objectFields[aoObject][-1].grid(row = aoObject + 1, column = aoColumn, sticky=N+E+S+W)

        for aoColumn in range(0, len(aoLabels) + 1):
            self.aoFrame.grid_columnconfigure(aoColumn,minsize = 100)

        presentations_label = Label(master, text="Presentations")
        presentations_label.pack(fill=X)

        self.presFrame = Frame(master, relief=RAISED, borderwidth=5)
        self.presFrame.pack(fill=BOTH, expand=True)

        presLabels = ["Name", "NameLang", "Bed Element", "Target Cfg", "Object Elements", "PresLang"]

        for presColumn in range(1, len(presLabels) + 1):
            presLabel = Label(self.presFrame, text = presLabels[presColumn - 1])
            presLabel.grid(row = 0, column = presColumn)

        self.presFields = [[] for i in range(self.numPresentations)]
        self.oeFrames = []
        self.objectElements = [[] for i in range(self.numObjects)]
        self.oeVars = [[] for i in range(self.numObjects)]
        self.playbackPresVar = IntVar(master)
        self.playbackPresVar.set(None)
        self.presButtons = []
        for presentation in range(0, self.numPresentations):
            self.presButtons.append(Radiobutton(self.presFrame, variable=self.playbackPresVar, state=DISABLED, command=self.playback_button, text=str(presentation + 1), indicatoron=0, value=presentation + 1))
            self.presButtons[-1].grid(row=presentation + 1, column=0, sticky=N+E+S+W)

            # First do regular indicator labels
            for presColumn in [1,2,3,4,6]:
                self.presFields[presentation].append(Label(self.presFrame, text = "None", relief=SUNKEN, bg = '#B3B4C8'))
                self.presFields[presentation][-1].grid(row = presentation + 1, column = presColumn, sticky=N+E+S+W)

            # Now indicators for object elements
            self.oeFrames.append(Frame(self.presFrame))
            self.oeFrames[-1].grid(row = presentation + 1, column = 5)
            for indicator in range(1,self.numObjects + 1):
                self.oeVars[presentation].append(IntVar())
                self.oeVars[presentation][indicator - 1].set(0);
                self.objectElements[presentation].append(Checkbutton(self.oeFrames[-1],state=DISABLED, variable = self.oeVars[presentation][indicator - 1]))
                self.objectElements[presentation][-1].pack(side=LEFT)

            for presColumn in range(0, len(presLabels) + 1):
                self.presFrame.grid_columnconfigure(presColumn, minsize = 100)

        button = Button(master, text="Run", command=self.run)
        button.pack(side=LEFT)


        button = Button(master, text="Quit", command=self.quit)
        button.pack(side=LEFT)

        if self.debug:
            button = Button(master, text="Memory", command=self.memoryDebug)
            button.pack(side=LEFT)

        # If a filename was specified then load it into UI, otherwise just wait for run to pressed to search for service
        if xml_filename is not None:
            if isFilePmd(xml_filename):
                try:
                    self.model = parse_pmd_xml(xml_filename, PMD_XML_MODE_FILE)
                    xml_filename.seek(0)
                except:
                    raise RuntimeError("Recognized file as PMD but parsing failed")
            elif isFileSADM(xml_filename):
                try:
                    self.model = populate_model_from_adm(xml_filename, ADM_XML_MODE_FILE)
                except:
                    raise RuntimeError("Recognized file as ADM but parsing failed")
            else:
                raise("File format not recognized")
            self.updateFromModel(self.model)
        else:
            # Create Drop down for IP Interfaces
            if platform.system() == 'Windows':
                fullIfListNames = [x['name'] for x in scapy.get_windows_if_list()]
                fullIfListGuids = [x['guid'] for x in scapy.get_windows_if_list()]
#				#full_ifList = netifaces.interfaces()
            else:
                fullIfListNames = scapy.get_if_list()

            # Filter interfaces according to those that have

            ifList = []
            if platform.system() == 'Windows':
                for iface in range(0,len(fullIfListNames)):
                    name = fullIfListNames[iface]
                    guid = fullIfListGuids[iface]
                    if netifaces.ifaddresses(guid).get(netifaces.AF_INET) != None and ("Loopback" not in name):
                        ifList.append(name)
            else:
                for iface in fullIfListNames:
                    if netifaces.ifaddresses(iface).get(netifaces.AF_INET) != None and iface != 'lo0':
                        ifList.append(iface)

            # Now we have a list of interfaces, create Drop-down menu to select interface
            self.interface = StringVar(master)
            if ifList is not []:
                self.interface.set(ifList[-1])
                self.ifList = OptionMenu(master, self.interface, *ifList, command=self.setInterfaceName)
                self.ifList.pack(side=LEFT)
            self.interfaceName = self.interface.get()

        label = Label(master, text="Service:")
        label.pack(side=LEFT)
        self.serviceSelection = StringVar(master)
        #self.serviceSelection.set(self.serviceNameList[0])
        self.serviceMenu = OptionMenu(master, self.serviceSelection, "Service")
        self.serviceMenu.pack(side=LEFT)
        label = Label(master, text="Audio:")
        label.pack(side=LEFT)
        self.audioSelection = StringVar(master)
        self.audioMenu = OptionMenu(master, self.audioSelection, "Audio")
        self.audioMenu.pack(side=LEFT)

        button = Button(master, text="XML", command=self.view_XML)
        button.pack(side=LEFT)

        # Now that the GUI is complete it is safe to start discovery
        # Option menu must exist first so this can't be moved up

        # Make sure a main thread exists so it is able to accept messages, even if it is not started

        self.messageQueue = queue.Queue()
        self.discoveryService = aoip_services.aoip_discovery.aoip_discovery(self.discoveryCallback)
        # If an SDP file was provided then add an additional service based on the file
        if sdp_filename is not None:
            self.discoveryService.add_aoip_service_from_sdp_file(sdp_filename)
        # This object is needed for joins and leaves
        self.multicast = aoip_services.multicast.MulticastGroup()

        try:
        	shellState = (platform.system() == "Windows")
        	self.haveGstreamer = ((subprocess.call(["gst-launch-1.0", "--version"], shell=shellState, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)) == 0)
        	print(self.haveGstreamer)
        except OSError as e:
        	if e.errno == errno.ENOENT:
        		self.haveGstreamer = False
        	else:
        		# Something else went wrong while trying to check for gstreamer
        		raise
        		self.haveGstreamer = False

    def post(self, command, data):
        self.messageQueue.put([command,data])

    def messageWaiting(self):
        return not self.messageQueue.empty()

    def getMessage(self):
        if self.messageQueue.empty():
            return None
        else:
            return self.messageQueue.get()

    def discoveryCallback(self, newService):
        self.post("newService", newService)

    def setInterfaceName(self, selected):
        self.interfaceName = self.interface.get()

    # This call provides a service that can be used to update the elements in the display
    # A task is created that sniffs packets and updates the parameters
    def set_service(self, service):
        # If we don't have a main thread then create one
        # If we do then signal it to stop
        self.multicast.leave()
        if self.mainThread is None:
            self.mainThread = StoppableThread(target = self.mainStart)
            self.service = service
            self.mainThread.start()
        else:
            self.service = service
            self.mainThread.reset()
        # The interface name that the multicast module needs has to be able to be used with sockets
        # which has a different format for Windows vs POSIX
        if platform.system() == "Windows":
            socketIfName = [x['guid'] for x in scapy.get_windows_if_list() if x['name'] == self.interfaceName][0]
        else:
            socketIfName = self.interfaceName
        self.multicast.join(socketIfName, self.service.sdp.address, self.service.sdp.port)

    def playback_button(self):
        if (self.playbackPres == self.playbackPresVar.get()):
            self.presButtons[self.playbackPres - 1].deselect()
            self.playbackPres = None
            # Stop playback here
        else:
            # Start playback here
            # First check to see if we are in a valid state to start playback
            # Check we are receiving metadata and an audio stream has been selected
            # This is enough to get started. Checking will happen as audio is brought up
            # If not then deselect
            # If the user has somehow managed to click a button that should be inactive then revert
            # this case should be unreachable
            if self.playbackPresVar.get() > len(self.model.audio_presentations):
                self.presButtons[self.playbackPresVar.get() - 1].deselect()
                if self.playbackPres is not None:
                    self.playbackPresVar.set(self.playbackPres)

            if self.indicators.rxMetadata() and self.audioSelection is not None:
                self.playbackPres = self.playbackPresVar.get()
                self.audioService = self.audioList[self.audioNameList.index(self.audioSelection.get())]
            else:
                self.presButtons[self.playbackPres - 1].deselect()
                self.playbackPres = None
                self.stop_playback()

    def pollPlayback(self):
        if self.playbackPres is not None:
            success = self.start_playback()
            # Playback can fail for many reasons, mostly invalid or changed metadata
            # Playback should stop and reach safe state
            if not success:
                #Not is main loop so can't deselect
                #self.presButtons[self.playbackPres - 1].deselect()
                self.stop_playback()
                self.playbackPres = None
        elif self.playing():
            self.stop_playback()


    def get_launch_str(self):
        # Now construct string to pass to gstreamer launch
        buffer_time = 0.6
        buffer_size = buffer_time * self.audioService.sdp.fs * self.audioService.sdp.channels
        if self.audioService.sdp.codec == 'L24':
            buffer_size = int(buffer_size * 3)
        else:
            buffer_size = int(buffer_size * 2)
        launchstr = 'udpsrc address=' + self.audioService.sdp.address + ' port=' + str(self.audioService.sdp.port) + \
                    ' buffer-size=' + str(buffer_size) + ' caps="application/x-rtp, media=(string)audio, clock-rate=(int)' + \
                    str(self.audioService.sdp.fs) + ', channels = ' + str(self.audioService.sdp.channels) + ', encoding-name=(string)' + \
                    self.audioService.sdp.codec + '" ! rtp' + self.audioService.sdp.codec + 'depay ! audio/x-raw,channels=' + \
                    str(self.audioService.sdp.channels) + ',channel-mask=(bitmask)0x000000000000 ! audioconvert mix-matrix="<<'
        for outchan in [0, 1]:
            for inchan in range(self.audioService.sdp.channels):
                launchstr = launchstr + '(float)' + str(self.mixMatrix[outchan][inchan]) + ', '
            launchstr = launchstr.rstrip(', ')
            launchstr = launchstr + '>, <'
        launchstr = launchstr.rstrip('>, <')
        launchstr = launchstr + '>>" ! playsink volume = 5 channels = 2'
        #launchstr = 'udpsrc address=239.150.138.1 port=5004 buffer-size=1000000 caps="application/x-rtp, media=(string)audio, clock-rate=(int)48000, channels = 9, encoding-name=(string)L16" ! rtpL16depay ! audio/x-raw,channels=9,channel-mask=(bitmask)0x000000000000 ! audioconvert mix-matrix="<<(float)1.0, (float)0.0, (float)1.0, (float)0.0, (float)1.0, (float)0.0, (float)1.0, (float)0.0, (float)0.0>, <(float)0.0, (float)1.0, (float)1.0, (float)0.0, (float)0.0, (float)1.0, (float)1.0, (float)0.0, (float)0.0>>" ! playsink volume = 5 channels = 2'
        return(launchstr)

    def start_playback(self):
        # Create mixing matrix by first adding in beds
        # Check that beds do not exceed total number of audio channels
        # Create empty mixing matrix

        # Stereo output only,
        newMixMatrix = [[0] * self.audioService.sdp.channels, [0] * self.audioService.sdp.channels]
        m3db = 0.7071
        bedCount = 0
        # If presentation was deleted then the playbackPres may be now invalid
        if self.playbackPres > len(self.model.audio_presentations):
            return False
        for element in self.model.audio_presentations[self.playbackPres - 1].elements:
            if type(element) is AudioBed and bedCount < self.numBeds:
                # We have a bed so mix
                startChannel = element.output_targets[0].audio_signals[0].id - 1
                endChannel = element.output_targets[0].audio_signals[0].id + \
                             PRESENTATION_CONFIG_EQUIV[PRESENTATION_CONFIG_TEXT.index(element.speaker_config)] - 2
                # Check that bed does not extend beyond audio service
                if endChannel <= startChannel or endChannel > self.audioService.sdp.channels:
                    return False
                if element.speaker_config == LOUDSPEAKER_CONFIG_COMMON_USE_TEXT_NAMES[LOUDSPEAKER_CONFIG_2_0]:
                    newMixMatrix[0][startChannel] = 1.0
                    newMixMatrix[1][startChannel + 1] = 1.0
                elif element.speaker_config == LOUDSPEAKER_CONFIG_COMMON_USE_TEXT_NAMES[LOUDSPEAKER_CONFIG_5_1]:
                    newMixMatrix[0][startChannel] = 1.0
                    newMixMatrix[1][startChannel + 1] = 1.0
                    newMixMatrix[0][startChannel + 2] = m3db
                    newMixMatrix[1][startChannel + 2] = m3db
                    newMixMatrix[0][startChannel + 4] = m3db
                    newMixMatrix[1][startChannel + 5] = m3db
                # unsupported bed configuration
                else:
                    return False
            if type(element) == AudioObject:
                # Check that audio object channel is not outside audio service
                if element.audio_signal.id > self.audioService.sdp.channels:
                    return False
                # Calculate Lo & Ro gains from X,Y,Z
                yz_prod = ((0.207 * element.elevation_or_y) + 0.5) * (m3db - (m3db * element.distance_or_z))
                newMixMatrix[0][element.audio_signal.id - 1] = (0.603 - (0.603 * element.azimuth_or_x)) * yz_prod * gain_from_db(element.source_gain_db)
                newMixMatrix[1][element.audio_signal.id - 1] = (0.603 + (0.603 * element.azimuth_or_x)) * yz_prod * gain_from_db(element.source_gain_db)

        # If pipeline is not playing or matrix has changed size then
        if not self.playing():
            self.mixMatrix = newMixMatrix
            launchstr = self.get_launch_str()
            self.launch_pipeline(launchstr)
        else:
            if len(newMixMatrix) != len(self.mixMatrix) or len(newMixMatrix[0]) != len(self.mixMatrix[0]):
                self.mixMatrix = newMixMatrix
                launchstr = self.get_launch_str()
                self.launch_pipeline(launchstr)
            else:
                # compare matrices
                diff = 0
                for row in range(len(newMixMatrix)):
                    for col in range(len(newMixMatrix[0])):
                        diff = diff + abs(newMixMatrix[row][col] - self.mixMatrix[row][col])
                if diff > 0.001:
                    self.mixMatrix = newMixMatrix
                    launchstr = self.get_launch_str()
                    self.launch_pipeline(launchstr)

        # Indicate that playback started successfully
        return True

    def launch_pipeline(self, launchstr):
        args = ['gst-launch-1.0', '-q']
        args = args + launchstr.split()
        shellState = (platform.system() == "Windows")
        newAudioPipeLine = subprocess.Popen(args, shell=shellState, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if self.audioPipeLine is not None:
            self.stop_playback()
        self.audioPipeLine = newAudioPipeLine

    def playing(self):
        return self.audioPipeLine is not None

    def stop_playback(self):
        if self.audioPipeLine is not None:
        	if (platform.system() == "Windows"):
        		subprocess.call(['taskkill', '/F', '/T', '/PID', str(self.audioPipeLine.pid)], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        	else:
        		os.kill(self.audioPipeLine.pid, signal.SIGSTOP)
        		os.kill(self.audioPipeLine.pid, signal.SIGKILL)
        self.audioPipeLine = None

    def reset_ui(self):
        # Reset lights
        self.indicators.reset()

        # Reset Metadata display

        for ununsedBeds in range(0,self.numBeds):
                self.bedFields[ununsedBeds][AudioBedHeadings.SOURCE].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.CONFIG].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.NAME].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.START].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.END].configure(text="")

        for unusedObject in range(0,self.numObjects):
                self.objectFields[unusedObject][AudioObjectHeadings.TYPE].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.NAME].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.DIVERGE].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.CH].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.GAIN].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.X].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.Y].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.Z].configure(text="")

        for unusedPresentations in range(0,self.numPresentations):
            self.presButtons[unusedPresentations]['state'] = 'disabled'
            self.presFields[unusedPresentations][PresentationHeadings.NAME].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.NAME_LANG].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.TARGET_CFG].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.PRES_LANG].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.BED_ELEMENT].configure(text="")

            for i in range(0,self.numObjects):
                self.oeVars[unusedPresentations][i].set(0)


    def quit(self):
        # Stop discovery
        self.multicast.leave()
        self.discoveryService.stop()
        self.playbackPres = None
        if self.playing():
            self.stop_playback()
        #shut down main thread if it is running
        if self.mainThread is not None and not self.mainThread.stopped():
            # Signal main Thread to Step
            self.mainThread.stop()
        else:
            self.post("quit", None)

    # Based on the selected service, set it as active
    def run(self):
        if len(self.serviceNameList) > 0:
            #service = self.serviceSelection.get()
            #name = self.serviceNameList.index(service)
            self.set_service(self.serviceList[self.serviceNameList.index(self.serviceSelection.get())])

    def memoryDebug(self):
        key_type='lineno'
        limit=10
        snapshot = tracemalloc.take_snapshot()
        snapshot = snapshot.filter_traces((tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),tracemalloc.Filter(False, "<unknown>")))
        top_stats = snapshot.statistics(key_type)
        print("Top %s lines" % limit)
        print('{: <10}'.format("index"),'{: <40}'.format("filename"), '{: <10}'.format("line no"),'{: <10}'.format("size kB"))
        for index, stat in enumerate(top_stats[:limit], 1):
            frame = stat.traceback[0]
            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
            print('{: <10}'.format(index),'{: <40}'.format(filename), '{: <10}'.format(frame.lineno),'{:0.1f}'.format(stat.size / 1024),"kB")
            line = linecache.getline(frame.filename, frame.lineno).strip()
            if line:
                print('    %s' % line)
        other = top_stats[limit:]
        if other:
            size = sum(stat.size for stat in other)
            print("%s other: %.1f KiB" % (len(other), size / 1024))
        total = sum(stat.size for stat in top_stats)
        print("Total allocated size: %.1f KiB" % (total / 1024))

    def view_XML(self):
        self.XMLViewerRequest = True

    def updateFromModel(self, model):
        objectCount = 0
        objectIdList = []
        bedCount = 0
        for element in model.audio_elements:
            if (type(element) is AudioObject) and (objectCount < self.numObjects):
                if element.classification < len(OBJECT_CLASSES):
                    self.objectFields[objectCount][AudioObjectHeadings.TYPE].configure(text=OBJECT_CLASSES[element.classification])
                self.objectFields[objectCount][AudioObjectHeadings.NAME].configure(text=element.name)
                if element.diverge == 1:
                    self.objectFields[objectCount][AudioObjectHeadings.DIVERGE].configure(text="True")
                else:
                    self.objectFields[objectCount][AudioObjectHeadings.DIVERGE].configure(text="False")
                if element.audio_signal is not None:
                    self.objectFields[objectCount][AudioObjectHeadings.CH].configure(text=element.audio_signal.id)
                else:
                    self.objectFields[objectCount][AudioObjectHeadings.CH].configure(text="")
                self.objectFields[objectCount][AudioObjectHeadings.GAIN].configure(text=element.source_gain_db)
                self.objectFields[objectCount][AudioObjectHeadings.X].configure(text=element.azimuth_or_x)
                self.objectFields[objectCount][AudioObjectHeadings.Y].configure(text=element.elevation_or_y)
                self.objectFields[objectCount][AudioObjectHeadings.Z].configure(text=element.distance_or_z)
                objectIdList.append(element.id)
                objectCount = objectCount + 1
            if type(element) is AudioBed and bedCount < self.numBeds:
                self.bedFields[bedCount][AudioBedHeadings.SOURCE].configure(text="Direct")
                self.bedFields[bedCount][AudioBedHeadings.CONFIG].configure(text=element.speaker_config)
                self.bedFields[bedCount][AudioBedHeadings.NAME].configure(text=element.name)
                self.bedFields[bedCount][AudioBedHeadings.START].configure(text=element.output_targets[0].audio_signals[0].id)
                self.bedFields[bedCount][AudioBedHeadings.END].configure(text=element.output_targets[0].audio_signals[0].id +
                    PRESENTATION_CONFIG_EQUIV[PRESENTATION_CONFIG_TEXT.index(element.speaker_config)] - 1)
                bedCount = bedCount + 1

        for ununsedBeds in range(bedCount, self.numBeds):
                self.bedFields[ununsedBeds][AudioBedHeadings.SOURCE].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.CONFIG].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.NAME].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.START].configure(text="")
                self.bedFields[ununsedBeds][AudioBedHeadings.END].configure(text="")

        for unusedObject in range(objectCount, self.numObjects):
                self.objectFields[unusedObject][AudioObjectHeadings.TYPE].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.NAME].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.DIVERGE].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.CH].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.GAIN].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.X].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.Y].configure(text="")
                self.objectFields[unusedObject][AudioObjectHeadings.Z].configure(text="")

        presentationCount = 0
        for presentation in model.audio_presentations:
            if presentationCount < self.numPresentations:
                if len(presentation.name_language) > 0:
                    self.presFields[presentationCount][PresentationHeadings.NAME].configure(text=presentation.name_language[0].name)
                    self.presFields[presentationCount][PresentationHeadings.NAME_LANG].configure(text=presentation.name_language[0].language)
                else:
                    self.presFields[presentationCount][PresentationHeadings.NAME].configure(text="")
                    self.presFields[presentationCount][PresentationHeadings.NAME_LANG].configure(text="")
                self.presFields[presentationCount][PresentationHeadings.TARGET_CFG].configure(text=presentation.config)
                self.presFields[presentationCount][PresentationHeadings.PRES_LANG].configure(text=presentation.language)
                objectIncludeList = [0] * self.numObjects
                for element in presentation.elements:
                    if type(element) == AudioObject:
                        objectIncludeList[objectIdList.index(element.id)] = 1
                    if type(element) == AudioBed:
                        self.presFields[presentationCount][PresentationHeadings.BED_ELEMENT].configure(text=element.name)
                for i in range(0, self.numObjects):
                    self.oeVars[presentationCount][i].set(objectIncludeList[i])
                if self.presButtons[presentationCount]['state'] == 'disabled' and self.haveGstreamer:
                    self.presButtons[presentationCount]['state'] = 'normal'
                presentationCount = presentationCount + 1

        for unusedPresentations in range(presentationCount, self.numPresentations):
            self.presButtons[unusedPresentations]['state'] = 'disabled'
            self.presFields[unusedPresentations][PresentationHeadings.NAME].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.NAME_LANG].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.TARGET_CFG].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.PRES_LANG].configure(text="")
            self.presFields[unusedPresentations][PresentationHeadings.BED_ELEMENT].configure(text="")

            for i in range(0, self.numObjects):
                self.oeVars[unusedPresentations][i].set(0)

    def processNextMessage(self, master):
        if self.messageWaiting():
            [command,messageData] = self.getMessage()
            if command == "newService":
                newService = messageData
                if newService.sdp.isAM824() or newService.sdp.isAESX242():
                    if self.debug:
                        print("Found ", newService.system, " Service ", newService.name)
                    self.serviceList.append(newService)
                    self.serviceNameList.append(newService.system + ':' + newService.name)
                    # Check if this is the first service, if so display it
                    if len(self.serviceNameList) == 1:
                        self.serviceSelection.set(self.serviceNameList[0])
                    # Now update GUI menu
                    self.serviceMenu['menu'].delete(0, 'end')
                    for service in self.serviceNameList:
                        self.serviceMenu['menu'].add_command(label=service, command=lambda value=service: self.serviceSelection.set(value))
                    # Set capture period according to received service type
                    if newService.sdp.isAESX242():
                        self.numCapturePackets = 1
                    else:
                        self.numCapturePackets = 42
                elif newService.sdp.isAES67():
                    self.audioList.append(newService)
                    self.audioNameList.append(newService.system + ':' + newService.name)
                    # Check if this is the first service, if so display it
                    if len(self.audioNameList) == 1:
                        self.audioSelection.set(self.audioNameList[0])
                    # Now update GUI menu
                    self.audioMenu['menu'].delete(0, 'end')
                    for service in self.audioNameList:
                        self.audioMenu['menu'].add_command(label=service, command=lambda value=service: self.audioSelection.set(value))

            if command == "reset":
                self.reset_ui()
            if command == "quit":
                master.quit()
                master.destroy()
                # Signal to main loop to quit
                return False
            if command == "PMD XML":
                XMLWindow = Toplevel(master)
                XMLWindow.geometry("1000x1000")
                XML_Viewer(XMLWindow, messageData, heading_text=" Audio Metadata Viewer  ").pack()
                self.XMLViewerRequest = False
            if command == "ADM XML":
                XMLWindow = Toplevel(master)
                XMLWindow.geometry("1500x1000")
                XML_Viewer(XMLWindow, messageData, heading_text=" Audio Metadata Viewer  ").pack()
                self.XMLViewerRequest = False
            if command == "updateModel":
                self.updateFromModel(messageData)
            if command == "updateInd":
                self.pmdInd.config(bg=self.indicators.pmd)
                self.sadmInd.config(bg=self.indicators.sadm)
                self.aesx242Ind.config(bg=self.indicators.aesx242)
                self.FrameInd.config(bg=self.indicators.frame)
                self.SubFrameInd.config(bg=self.indicators.subframe)
        return True

    def mainStart(self):
        while not self.mainThread.stopped():
            deframer = pmdDeframer()
            self.post("reset", None)
            self.mainThread.clear_reset()

            while not (self.mainThread.is_reset() or self.mainThread.stopped()):
                    try:
                        packets = scapy.sniff(iface=self.interfaceName, filter="ip dst " + self.service.sdp.address + " and udp dst port " + str(self.service.sdp.port), count=self.numCapturePackets, timeout=1)
                    except RuntimeError as e:
                        print(e)
                        packets = []
                        print("Error capturing packets...Stopping Capture")
                        self.mainThread.stop();
                    if len(packets) > 0:
                        for packet in packets:
                            if packet["UDP"].dport == self.service.sdp.port:
                                deframer.receivePacket(packet, self.service.sdp.codec)
                            if deframer.haveFrame():
                                break
                        if deframer.haveFrame():
                            newFrame = deframer.getFrame()
                            try:
                                if newFrame.is_pmd():
                                    if get_pmd_xml_from_wav(newFrame):
                                        if (check_xml_complete()):
                                            try:
                                                self.model = parse_pmd_xml("pmd.xml", PMD_XML_MODE_FILE)
                                            except:
                                                self.indicators.pmdError()
                                                raise RuntimeError("PMD XML parsing failed!")
                                            self.post("updateModel", copy.deepcopy(self.model))
                                            self.indicators.pmdOn()
                                    else:
                                        self.indicators.pmdError()
                                        raise RuntimeError("PMD conversion to XML failed!")
                                elif newFrame.is_sADM():
                                    xmlText = get_sadm_xml(newFrame)
                                    if xmlText is not None:
                                        try:
                                            #model = populate_model_from_adm("sadm.xml", ADM_XML_MODE_FILE)
                                            self.model = populate_model_from_adm(xmlText, ADM_XML_MODE_STRING)
                                        except:
                                            self.indicators.sadmError()
                                            raise RuntimeError("SADM XML parsing failed")
                                        self.post("updateModel", copy.deepcopy(self.model))
                                        self.indicators.sadmOn()
                                    else:
                                        self.indicators.sadmError()
                                        raise RuntimeError("SADM conversion to XML failed!")
                                elif newFrame.is_AESX242():
                                    if get_pmd_xml_from_klv(newFrame,False):
                                        try:
                                            self.model = parse_pmd_xml("pmd.xml", PMD_XML_MODE_FILE)
                                        except:
                                            self.indicators.aesx242Error()
                                            raise RuntimeError("AESX242 XML parsing failed!")
                                        self.post("updateModel", copy.deepcopy(self.model))
                                        self.indicators.aesx242On()
                                    else:
                                        self.indicators.aesx242Error()
                                        raise RuntimeError("AESX242 conversion to XML failed!")
                                else:
                                    # received some unknown format
                                    self.indicators.reset()

                                # Handle framing indicators for PMD & SADM
                                if newFrame.is_pmd() or newFrame.is_sADM():
                                    if newFrame.subframe_mode:
                                        self.indicators.subFrameMode()
                                    else:
                                        self.indicators.frameMode()

                                if (self.XMLViewerRequest):
                                    if newFrame.is_pmd() or newFrame.is_AESX242():
                                        with open("pmd.xml", "r") as xmlFile:
                                            xmlText = xmlFile.read()
                                        self.post("PMD XML", xmlText)
                                    else:
                                        self.post("ADM XML", xmlText)
                                    self.XMLViewerRequest = False


                            except RuntimeError as e:
                                print(e)
                    else:
                        # No packets received before timeout
                        self.post("reset", None)
                    if self.debug:
                        # fps counter, useful for establishing performance
                        print("fps:", round(1/(time.time() - self.start_time)), sep=' ', end='  \r', flush=True)
                        self.start_time = time.time()
                    self.pollPlayback()
     #   			scapy.hexdump(payload24)
    #                for pkt in pkts:
    #                   print(pkt["RTP"].sequence)
        sys.stdout.flush()
        # Quit
        clean_up_temp_files(["pmd.klv", "pmd.wav", "pmd.xml"])
        self.post("quit", None)

# The main thread takes a single file as input. This takes a single filename
# If a filename is provided then the UI prepolulates with that file
# If not then the UI starts empty and waits for control input with the UI gadgets

def main():
    parser = argparse.ArgumentParser(description='Audio Metadata (AM) realtime viewer application')
    parser.add_argument('-xml', type=argparse.FileType('r'), default=None, help='XML filename for offline display of XML')
    parser.add_argument('-sdp', type=argparse.FileType('r'), default=None, help='SDP filename to avoid requiring stream discovery')
    parser.add_argument('-debug', action='store_const', const=True, help='Enables debug mode')
    args = parser.parse_args()

    if args.debug:
        global tracemalloc
        import tracemalloc
        global linecache
        import linecache
        tracemalloc.start()

    root = Tk()
    root['bg'] = '#B3B4D8'
    root.tk_setPalette(background='#A3A4C8', foreground='black',
                   activeBackground='#A3A4C8', activeForeground='black')
    my_gui = PmdAdmDisplayGUI(root, args.xml, args.sdp, args.debug)

    root.update()

    while my_gui.processNextMessage(root):
        root.update()

#	root.mainloop()

if __name__ == '__main__':
    print("Starting...")
    main()



#!/usr/bin/env python

"""
 AOIP Services
 Copyright (c) 2019, Dolby Laboratories Inc.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted
 provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
    and the following disclaimer in the documentation and/or other materials provided with the distribution.
 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or
    promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
"""

class sdp_parser:

    type = ""
    codec = ""
    fs = 0
    address = ""
    port = 0
    session_name = ""
    channels = 0

    def __init__(self, sdp_text):
        for line in sdp_text.splitlines():
            split_line = line.split('=',1)
            if len(split_line) > 1:
                header = split_line[0]
                body = split_line[1].split(' ')
                if header == 's' and len(body) > 0:
                    self.session_name = body[0]
                if header == 'm' and len(body) > 1:
                        self.type = body[0]
                        self.port = int(body[1])
                if header == 'c' and len(body) > 0:
                    self.address = body[-1].split('/',1)[0]
                if header == 'a' and len(body) > 1:
                    if body[0].find("rtpmap:") != -1:
                        format = body[1].split('/',2)
                        if len(format) > 1:
                            self.codec = format[0]
                            self.fs = int(format[1])
                            if len(format) > 2:
                                self.channels = int(format[2])
                            else:
                                self.channels = 0

    def print(self):
        print("SDP Contents")
        print("============")
        print("Name:", self.session_name)
        print("Type: ", self.type)
        print("Multicast address: ", self.address)
        print("Port", self.port)
        print("Codec: ", self.codec)
        print("Sampling Frequency: ", self.fs,"Hz")
        print("No of channels:", self.channels)


    def isAM824(self):
        return((self.codec == "AM824") and (self.fs == 48000) and (self.channels == 2))

    def isAESX242(self):
        return(self.codec == "smpte336m")

    def isAES67(self):
        return(self.codec == "L16" or self.codec == "L24")


"""
****************************************************************************************************************************************************************
****************************************************************************************************************************************************************

    ___    ____  __  ___            ____  __  _______     ____               _           __
   /   |  / __ \/  |/  /           / __ \/  |/  / __ \   / __ \_________    (_)__  _____/ /_
  / /| | / / / / /|_/ /  ______   / /_/ / /|_/ / / / /  / /_/ / ___/ __ \  / / _ \/ ___/ __/
 / ___ |/ /_/ / /  / /  /_____/  / ____/ /  / / /_/ /  / ____/ /  / /_/ / / /  __/ /__/ /_
/_/  |_/_____/_/  /_/           /_/   /_/  /_/_____/  /_/   /_/   \____/_/ /\___/\___/\__/
                                                                      /___/

****************************************************************************************************************************************************************
****************************************************************************************************************************************************************

Copyright (c) 2018, Dolby Laboratories Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

****************************************************************************************************************************************************************
****************************************************************************************************************************************************************
"""

# ************************************************************************************************************************************************************ #
# ************************************************************************************************************************************************************ #

from adm.adm_const import NON_DIALOGUE_CONTENT_KIND
from adm.adm_const import DIALOGUE_CONTENT_KIND
from adm.adm_const import MIXED_CONTENT_KIND

from adm.adm_const import DIALOGUE_CONTENT_KIND
from adm.adm_const import NON_DIALOGUE_CONTENT_KIND
from adm.adm_const import MIXED_CONTENT_KIND
from adm.adm_const import DIALOGUE_CONTENT
from adm.adm_const import NON_DIALOGUE_CONTENT
from adm.adm_const import MIXED_CONTENT
from adm.adm_const import DIALOGUE_CONTENT
from adm.adm_const import DIALOGUE
from adm.adm_const import VOICEOVER
from adm.adm_const import SPOKEN_SUBTITLE
from adm.adm_const import AUDIO_DESCRIPTION
from adm.adm_const import COMMENTARY
from adm.adm_const import EMERGENCY
from adm.adm_const import TYPE_DEFINITION
from adm.adm_const import TYPE_LABEL

from adm.adm_const import AUDIO_PACK_STR
from adm.adm_const import AUDIO_PROGRAMME_STR
from adm.adm_const import AUDIO_CONTENT_STR
from adm.adm_const import AUDIO_OBJECT_STR
from adm.adm_const import AUDIO_BLOCK_STR
from adm.adm_const import AUDIO_CHANNEL_STR
from adm.adm_const import AUDIO_TRACK_UID_STR

from adm.adm_const import DIRECTSPEAKERS
from adm.adm_const import MATRIX
from adm.adm_const import OBJECTS
from adm.adm_const import HOA
from adm.adm_const import BINAURAL
from adm.adm_const import POLAR
from adm.adm_const import CARTESIAN

from adm.adm_const import DEFAULT_AUDIO_SAMPLE_RATE
from adm.adm_const import DEFAULT_AUDIO_BIT_DEPTH

# ************************************************************************************************************************************************************ #


class AudioFormatExtended(object):
    def __init__(self):
        self.audio_programme = []
        self.audio_content = []
        self.audio_object = []
        self.audio_pack_format = []
        self.audio_channel_format = []
        self.audio_track_uid = []


class LoudnessMetadata:
    def __init__(self):
        self.loudness_method = ""
        self.loudness_rec_type = ""
        self.loudness_correction_type = ""
        self.integrated_loudness = 0.0
        self.loudness_range = 0.0
        self.max_true_peak = 0.0
        self.max_momentary = 0.0
        self.max_short_term = 0.0
        self.dialogue_loudness = 0.0

    def update_loudness_method(self, loudness_method):
        self.loudness_method = loudness_method

    def update_loudness_rec_type(self, loudness_rec_type):
        self.loudness_rec_type = loudness_rec_type

    def update_loudness_correction_type(self, loudness_correction_type):
        self.loudness_correction_type = loudness_correction_type

    def update_integrated_loudness(self, integrated_loudness):
        self.integrated_loudness = integrated_loudness

    def update_loudness_range(self, loudness_range):
        self.loudness_range = loudness_range

    def update_max_true_peak(self, max_true_peak):
        self.max_true_peak = max_true_peak

    def update_max_momentary(self, max_momentary):
        self.max_momentary = max_momentary

    def update_max_short_term(self, max_short_term):
        self.max_short_term = max_short_term

    def update_dialogue_loudness(self, dialogue_loudness):
        self.dialogue_loudness = dialogue_loudness


class AudioProgrammeLabel:
    def __init__(self, name, language):
        self.name = name
        self.language = language


class AudioPackFormat:
    def __init__(self, id, name, type_label):
        self.id = id
        self.name = name
        self.audio_channel_idref = []
        self.type_label = type_label


class AudioProgramme:
    def __init__(self, name, language, id):
        self.name = name
        self.language = language
        self.id = id
        self.audio_programme_label = []
        self.audio_content_idref = []
        self.loudness_metadata = None


class AudioContent:
    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.dialogue = None
        self.audio_object_idref = []


class Dialogue:
    def __init__(self, value, classification):
        value = int(value)
        classification = int(classification)
        self.value = value
        if value == NON_DIALOGUE_CONTENT:
            self.content_kind = NonDialogueContentKind(classification)
        elif value == DIALOGUE_CONTENT:
            self.content_kind = DialogueContentKind(classification)
        elif value == MIXED_CONTENT:
            self.content_kind = MixedContentKind(classification)


class NonDialogueContentKind:
    def __init__(self, classification):
        self.classification = classification
        self.classification_text = NON_DIALOGUE_CONTENT_KIND[classification]


class DialogueContentKind:
    def __init__(self, classification):
        self.classification = classification
        self.classification_text = DIALOGUE_CONTENT_KIND[classification]


class MixedContentKind:
    def __init__(self, classification):
        self.classification = classification
        self.classification_text = MIXED_CONTENT_KIND[classification]


class AudioObject:
    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.audio_pack_idref = []
        self.audio_track_idref = []
        self.head_locked = None
        self.gain = None


class AudioChannelFormat:
    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.audio_block = None


class AudioBlockFormat:
    def __init__(self, id):
        self.id = id
        self.cartesian = None
        self.position_coord = None
        self.headphone_render = None
        self.speaker_label = None


class Position:
    def __init__(self, is_polar):
        self.is_polar = is_polar
        self.x_or_az = None
        self.y_or_el = None
        self.z_or_ds = None

class Headphone:
    def __init__(self, value, bypass, mode):
        self.value = value
        self.bypass = bypass
        self.mode = mode


class AudioTrackUID:
    def __init__(self, id):
        self.id = id
        self.channel_format_id = None
        self.pack_format_id = None


"""

class AudioContent:
    def __init__(self, id, name):
        self.id = id
        self.name = name


class AudioProgrammeLabel:
    def __init__(self, value, language):
        self.language = language
        self.name = value

    def update_language(self, language):
        self.language = language

    def update_name(self, name):
        self.name = name
        
        
class AudioPackFormat:
    def __init__(self, id, type_definition, name):

        # Note that custom id's start at 1000 not 0000
        if id < 0x10:
            idhext = "100"
        elif id < 0x100:
            idhext = "10"
        elif id < 0x1000:
            idhext = "1"
        elif id < 0x10000:
            idhext = ""

        # ID for the object is based upon its type_definition
        self.id = AUDIO_PACK_STR + str(type_definition) + idhext + hex(id)[2:]
        self.name = name
        self.type_label = TYPE_LABEL[type_definition]
        self.type_definition = TYPE_DEFINITION[type_definition]
        self.importance = 0
        self.audio_channel_format_id_refs = []
        self.audio_pack_format_id_refs = []
        self.absolute_distance = 0.0

    def update_id(self, id):
        self.id = id

    def update_name(self, name):
        self.name = name

    def update_type_label(self, type_label):
        self.type_label = type_label

    def update_type_definition(self, type_definition):
        self.type_definition = type_definition

    def update_importance(self, importance):
        self.importance = importance

    def update_absolute_distance(self, absolute_distance):
        self.absolute_distance = absolute_distance
class AudioProgramme:
    def __init__(self, name, language, id):
        if id < 0x10:
            idhext = "000"
        elif id < 0x100:
            idhext = "00"
        elif id < 0x1000:
            idhext = "0"
        elif id < 0x10000:
            idhext = ""

        self.id = AUDIO_PROGRAMME_STR + idhext + hex(id)[2:]
        self.name = name
        self.language = language
        self.start = ""
        self.end = ""
        self.max_ducking_depth = ""
        self.audio_program_labels = []
        self.content_id_refs = []
        self.loudness = []

    def update_id(self, id):
        self.id = id

    def update_name(self, name):
        self.name = name

    def update_language(self, language):
        self.language = language

    def update_start_time(self, start):
        self.start = start

    def update_end_time(self, end):
        self.end = end

    def update_max_ducking_depth(self, max_ducking_depth):
        self.max_ducking_depth = max_ducking_depth

    def add_programme_label(self, value, language):
        self.audio_program_labels.append(AudioProgrammeLabel(value, language))

    def add_programme_loudness(self):
        self.loudness.append(LoudnessMetadata())
"""


"""
class AudioContent:
    def __init__(self, id, name):
        if id < 0x10:
            idhext = "000"
        elif id < 0x100:
            idhext = "00"
        elif id < 0x1000:
            idhext = "0"
        elif id < 0x10000:
            idhext = ""

        self.id = AUDIO_CONTENT_STR + idhext + hex(id)[2:]
        self.name = name
        self.language = ""
        self.dialogue = Dialogue()
        self.audio_object_id_refs = []

    def update_id(self, id):
        self.id = id

    def update_name(self, name):
        self.name = name

    def update_language(self, language):
        self.language = language

    def update_dialogue(self, dialog_type):
        self.dialogue.update_value(DIALOGUE_CONTENT, dialog_type)

    def update_non_dialogue(self, non_dialogue_type):
        self.dialogue.update_value(NON_DIALOGUE_CONTENT, non_dialogue_type)


class Dialogue:
    def __init__(self):
        self.value = 0
        self.content_kind = NonDialogueContentKind()

    def update_value(self, value, classification, kind):
        self.value = value
        if value == NON_DIALOGUE_CONTENT:
            self.content_kind = NonDialogueContentKind()
        elif value == DIALOGUE_CONTENT:
            self.content_kind = DialogueContentKind()
        elif value == MIXED_CONTENT:
            self.content_kind = MixedContentKind()

        self.content_kind.update_classification(classification)
"""




"""
class AudioObject:
    def __init__(self, id, name):
        if id < 0x10:
            idhext = "000"
        elif id < 0x100:
            idhext = "00"
        elif id < 0x1000:
            idhext = "0"
        elif id < 0x10000:
            idhext = ""

        self.id = AUDIO_OBJECT_STR + idhext + hex(id)[2:]
        self.name = name
        self.start = ""
        self.duration = ""
        self.dialogue = 0
        self.importance = 0
        self.interact = 0
        self.disable_ducking = 0
        self.audio_pack_format_id_refs = []
        self.audio_object_id_refs = []
        self.audio_complimentary_object_id_refs = []
        self.audio_track_uid_refs = []
        self.object_interaction = AudioObjectInteraction()
        self.gain = 1.0
        self.headlocked = 0

    def update_id(self, id):
        self.id = id

    def update_name(self, name):
        self.name = name

    def update_start(self, start):
        self.start = start

    def update_duration(self, duration):
        self.duration = duration

    def update_dialogue(self, dialogue):
        self.dialogue = dialogue

    def update_importance(self, importance):
        self.importance = importance

    def update_interact(self, interact):
        self.interact = interact

    def update_disable_ducking(self, disable_ducking):
        self.disable_ducking = disable_ducking

    def update_gain(self, gain):
        self.gain = gain

    def update_headlocked(self, headlocked):
        self.headlocked = headlocked
"""


class ItemValidationData:
    def __init__(self, min_occurs, max_occurs):
        self.min_occurs = min_occurs
        self.max_occurs = max_occurs


class AudioObjectInteraction:
    def __init__(self):
        self.on_off_interact = 0
        self.gain_interact = 0
        self.position_interact = 0
        self.gain_interaction_range = GainInteractionRange()
        self.position_interaction_range = PositionInteractionRange()

    def update_interactive(self, on_off_interact, gain_interact, position_interact):
        self.on_off_interact = on_off_interact
        self.gain_interact = gain_interact
        self.position_interact = position_interact


class GainInteractionRange:
    def __init__(self):
        self.gain_min = 1.0
        self.gain_max = 1.0

    def update_min_max_gain(self, gain_min, gain_max):
        self.gain_min = gain_min
        self.gain_max = gain_max


class PositionInteractionRange:
    def __init__(self):
        self.azimuth_or_x_min = 0.0
        self.azimuth_or_x_max = 0.0
        self.elevation_or_y_min = 0.0
        self.elevation_or_y_max = 0.0
        self.distance_or_z_min = 0.0
        self.distance_or_z_max = 0.0

    def update_position_min_max(self, azimuth_or_x_min, azimuth_or_x_max, elevation_or_y_min, elevation_or_y_max, distance_or_z_min, distance_or_z_max):
        self.azimuth_or_x_min = azimuth_or_x_min
        self.azimuth_or_x_max = azimuth_or_x_max
        self.elevation_or_y_min = elevation_or_y_min
        self.elevation_or_y_max = elevation_or_y_max
        self.distance_or_z_min = distance_or_z_min
        self.distance_or_z_max = distance_or_z_max


"""
class AudioBlockFormat:
    def __init__(self, id, type_definition, coordinate_mode, update_number):
        if id < 0x10:
            idhext = "100"
        elif id < 0x100:
            idhext = "10"
        elif id < 0x1000:
            idhext = "1"
        elif id < 0x10000:
            idhext = ""

        str_update_number = hex(update_number)[2:]
        i = 8 - len(str_update_number)
        while i > 0:
            str_update_number = "0" + str_update_number
            i -= 1

        self.id = AUDIO_BLOCK_STR + str(type_definition) + idhext + hex(id)[2:] + "_" + str_update_number
        self.rtime = ""
        self.duration = ""
        self.gain = 1.0
        self.coordinate_mode = coordinate_mode

        if type_definition == DIRECTSPEAKERS:
            self.definition = DirectSpeakers()
        elif type_definition == MATRIX:
            self.definition = Matrix()
        elif type_definition == OBJECTS:
            self.definition = Objects(self.coordinate_mode)
        elif type_definition == HOA:
            self.defination = HoA()
        elif type_definition == BINAURAL:
            self.definition = Binaural()

        if self.coordinate_mode == POLAR:
            self.azimuth = 0.0
            self.elevation = 0.0
            self.distance = 0.0
        elif self.coordinate_mode == CARTESIAN:
            self.x = 0.0
            self.y = 0.0
            self.z = 0.0

    def update_id(self, id):
        self.id = id

    def update_rtime(self, rtime):
        self.rtime = rtime

    def update_duration(self, duration):
        self.duration = duration

    def update_gain(self, gain):
        self.gain = gain

    def update_type_definition(self, typedef):
        if typedef == DIRECTSPEAKERS:
            self.definition = DirectSpeakers()
        elif typedef == MATRIX:
            self.definition = Matrix()
        elif typedef == OBJECTS:
            self.definition = Objects(self.coordinate_mode)
        elif typedef == HOA:
            self.defination = HoA()
        elif typedef == BINAURAL:
            self.definition = Binaural()

    def update_coordinates(self, azimuth_or_x, elevation_or_y, distance_or_z):
        if self.coordinate_mode == POLAR:
            self.azimuth = azimuth_or_x
            self.elevation = elevation_or_y
            self.distance = distance_or_z
        elif self.coordinate_mode == CARTESIAN:
            self.x = azimuth_or_x
            self.y = elevation_or_y
            self.z = distance_or_z




"""


class DirectSpeakers:
    def __init__(self,):
        self.speaker_label = ""

    def update_speaker_label(self, speaker_label):
        self.speaker_label = speaker_label


class Objects:
    def __init__(self, coordinate_mode):
        self.cartesian = coordinate_mode
        self.width = 0.0
        self.depth = 0.0
        self.height = 0.0
        self.diffuse = 0
        self.channel_lock = ChannelLock()
        self.divergence_range = 0.0
        self.jump_position = JumpPosition()
        self.zone_exclusion = []


class Zone:
    def __init__(self, name, min_azimuth_or_x, max_azimuth_or_x, min_elevation_or_y, max_elevation_or_y, min_z, max_z):
        self.name = name
        self.min_azimuth_or_x = min_azimuth_or_x
        self.max_azimuth_or_x = max_azimuth_or_x
        self.min_elevation_or_y = min_elevation_or_y
        self.max_elevation_or_y = max_elevation_or_y
        self.min_z = min_z
        self.max_z = max_z

    def update_zone_position(self, min_azimuth_or_x, max_azimuth_or_x, min_elevation_or_y, max_elevation_or_y, min_z, max_z):
        self.min_azimuth_or_x = min_azimuth_or_x
        self.max_azimuth_or_x = max_azimuth_or_x
        self.min_elevation_or_y = min_elevation_or_y
        self.max_elevation_or_y = max_elevation_or_y
        self.min_z = min_z
        self.max_z = max_z


class ChannelLock:
    def __init__(self):
        self.value = 0
        self.max_distance = 0.0

    def update(self, value, max_distance):
        self.value = value
        self.max_distance = max_distance


class JumpPosition:
    def __init__(self):
        self.value = 0
        self.interpolation_length = 0.0

    def update(self, value, interpolation_length):
        self.value = value
        self.interpolation_length = interpolation_length


class HeadphoneRender:
    def __init__(self):
        self.bypass = 0
        self.mode = 0

    def update(self, bypass, mode):
        self.bypass = bypass
        self.mode = mode

"""
class AudioChannelFormat:
    def __init__(self, id, type_definition, name):
        if id < 0x10:
            idhext = "000"
        elif id < 0x100:
            idhext = "00"
        elif id < 0x1000:
            idhext = "0"
        elif id < 0x10000:
            idhext = ""

        self.id = AUDIO_CHANNEL_STR + str(type_definition) + idhext + hex(id)[2:]
        self.name = name
        self.type_label = TYPE_LABEL[type_definition]
        self.type_definition = TYPE_DEFINITION[type_definition]
        self.frequency = []
        self.audio_block_format_id_ref = []

    def update_name(self, name):
        self.name = name

    def update_id(self, id):
        self.id = id

    def update_type_label(self, type_label):
        self.type_label = type_label

    def update_type_definition(self, type_definition):
        self.type_definition = type_definition
"""


"""
class AudioTrackUID:
    def __init__(self, id):
        if id < 0x10:
            idhext = "000"
        elif id < 0x100:
            idhext = "00"
        elif id < 0x1000:
            idhext = "0"
        elif id < 0x10000:
            idhext = ""

        self.id = AUDIO_TRACK_UID_STR + idhext + hex(id)[2:]
        self.sample_rate = DEFAULT_AUDIO_SAMPLE_RATE
        self.bit_depth = DEFAULT_AUDIO_BIT_DEPTH
        self.audio_mxf_lookup = AudioMXFLookUp()
        self.audio_track_format_id_ref = []
        self.audio_channel_format_id_ref = []
        self.audio_pack_format_id_ref = []

"""


class AudioMXFLookUp:
    def __init__(self):
        self.umid = ""
        self.track_id_ref = 0
        self.channel_id_ref = 0

    def update_audio_mxf_lookup(self, umid, track_id_ref, channel_id_ref):
        self.umid = umid
        self.track_id_ref = track_id_ref
        self.channel_id_ref = channel_id_ref


class AudioTrackFormat:
    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.format_label = ""
        self.format_definition = ""
        self.audio_stream_format_id_ref = []

    def update_id(self, id):
        self.id = id

    def update_name(self, name):
        self.name = name

    def update_format_label(self, format_label):
        self.format_label = format_label

    def update_format_definition(self, format_definition):
        self.format_definition = format_definition


class AudioStreamFormat:
    def __init__(self, id, name):
        self.id = id
        self.name = name
        self.format_label = ""
        self.format_definition = ""
        self.audio_channel_format_id_ref = []
        self.audio_pack_format_id_ref = []
        self.audio_track_format_id_ref = []

    def update_id(self, id):
        self.id = id

    def update_name(self, name):
        self.name = name

    def update_format_label(self, format_label):
        self.format_label = format_label

    def update_format_definition(self, format_definition):
        self.format_definition = format_definition

# ************************************************************************************************************************************************************ #
# ************************************************************************************************************************************************************ #



"""
 AOIP Services
 Copyright (c) 2019, Dolby Laboratories Inc.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted
 provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
    and the following disclaimer in the documentation and/or other materials provided with the distribution.
 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or
    promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import logging
import socket
import sys
import platform
import struct
from time import sleep
from typing import cast

from zeroconf import ServiceBrowser, ServiceStateChange, Zeroconf
import socket
import threading
import aoip_services.sdp_parser
from tkinter import *
import errno

class StoppableThread(threading.Thread):

	def __init__(self, **kwargs):
		super(StoppableThread, self).__init__(**kwargs)
		self._stop_event = threading.Event()
		self._lock = threading.Lock()

	def stop(self):
		self._stop_event.set()

	def stopped(self):
		return self._stop_event.is_set()

	def get_lock(self):
		self._lock.acquire()

	def release_lock(self):
		self._lock.release()


class SAPPacket:
	def __init__(self, data, addr):
#		scapy.hexdump(data)
		self.Version = data[0] >> 5;
		self.IsIPv6 = bool((data[0] >> 4) & 1)
		self.T = (data[0] >> 2) & 1
		self.IsEncrypted = bool((data[0] >> 1) & 1)
		self.IsCompressed = bool(data[0] & 1)
		self.AuthLen = int(data[1])
		self.MsgIdHash = struct.unpack("H", data[2:4])
		data = data[4:]
		if self.IsIPv6:
			self.OrigSource = data[:4*4]
			data = data[4*4:]
		else:
			self.OrigSource = data[:4]
			data = data[4:]
		# ignore AuthData
		data = data[self.AuthLen * 4:]
		if self.IsEncrypted:
			# ignore timestamp for now...
			data = data[4:]
		# search 0-byte, set payload type
		self.PayloadType = None
		for i in range(0, len(data)):
			if data[i] == 0:
				self.PayloadType = data[0:i]
				data = data[i+1:]
				break
		self.PayloadData = data
		
		try:
			mediaAddr = self.Payload.c.split()[-1].split("/")[0]
		except:
			mediaAddr = None
		try:
			mediaPort = self.Payload.m.split()[1]
		except:
			mediaPort = None
		if mediaAddr and mediaPort:
			self.Media = {}
			self.Media["Sender"] = self.Payload.s
			self.Media["Playgroup"] = self.Payload.a.get("x-plgroup", "")
			self.Media["Addr"] = "udp://@" + mediaAddr + ":" + mediaPort

class aoip_service:

	name = ""
	system = None
	port = ""
	ip_address = ""
	codec = ""
	sdp_text = ""
	sdp = None

class aoip_discovery:

	service_list = []
	service_list_lock = None
	callback = None
	dante_task = None

# call back is a function called when a service is added to the service list
# this is call
	def __init__(self, callback):
		self.callback = callback

		# Startup ravenna discovery
		zeroconf = Zeroconf()
		browser = ServiceBrowser(zeroconf, "_ravenna_session._sub._rtsp._tcp.local.", handlers=[self.found_ravenna_service])
 
		# Startup Dante discovery
		self.dante_task = StoppableThread(target = self.dante_discovery_task)
		self.dante_task.start()
		self.service_list_lock = threading.Lock()


	def stop(self):
		self.dante_task.stop()

	def dante_discovery_task(self):

		sap_addr = "239.255.255.255"
		port = 9875

		sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

		# Look up multicast group address in name server and find out IP version
		addrinfo = socket.getaddrinfo(sap_addr, None)[0]

		# Multicast code from: http://wiki.python.org/moin/UdpCommunication

		# Set some options to make it multicast-friendly
		sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		try:
			sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
		except AttributeError:
			pass # Some systems don't support SO_REUSEPORT
		sock.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_TTL, 20)
		sock.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_LOOP, 1)

		if addrinfo[0] == socket.AF_INET:
			membership_request = struct.pack("4sl", socket.inet_aton(sap_addr), socket.INADDR_ANY)
			# See http://www.tldp.org/HOWTO/Multicast-HOWTO-6.html for explanation of sockopts
			sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, membership_request)
		else:
			membership_request = struct.pack("16s16s", socket.inet_pton(socket.AF_INET6,sap_addr), chr(0)*16)
			sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, membership_request)

		if platform.system() == 'Windows':
			sock.bind( ('', port) ) # standard SAP port
		else:
			sock.bind( (sap_addr, port) ) # standard SAP port

		group_bin = socket.inet_pton(addrinfo[0], addrinfo[4][0])

		sock.setblocking(False)

		while not self.dante_task.stopped():
			try:
				data, addr = sock.recvfrom(1024)
			except socket.error as e:
				err = e.args[0]
				if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
					sleep(1)
					continue
				else:
					raise e

#			print ("received:", data, addr)
#			scapy.hexdump(data)
			if data:
				p = SAPPacket(data,addr)
#			print(p.PayloadData.decode())
				self.add_aoip_service_from_sdp_text("SAP", p.PayloadData.decode())

		# Leave SAP group
		membership_request = struct.pack("4sl", socket.inet_aton(sap_addr), socket.INADDR_ANY)
		sock.setsockopt(socket.IPPROTO_IP, socket.IP_DROP_MEMBERSHIP, membership_request)

		sock.close()



# Create a service based on an SDP text string
# This enables external discovery via shared SDP files
	def add_aoip_service_from_sdp_text(self,system,sdp_text):
		new_service = aoip_service()
		new_service.sdp = aoip_services.sdp_parser.sdp_parser(sdp_text)
		new_service.system = system
		new_service.name = new_service.sdp.session_name
		new_service.port = new_service.sdp.port
		new_service.ip_address = new_service.sdp.address
		new_service.codec = new_service.sdp.codec
		found = False
		for service in self.service_list:
			if new_service.name == service.name and new_service.system == service.system:
				found = True
		if not found:
			# This function can be called from multiple threads so this ensures that
			# list additions happen sequentially
			self.service_list_lock.acquire()
			self.service_list.append(new_service)
			self.service_list_lock.release()
			self.callback(new_service)

	def add_aoip_service_from_sdp_file(self,sdp_file):
		sdp_text=sdp_file.read()
		self.add_aoip_service_from_sdp_text("SDP",sdp_text)


	def found_ravenna_service(
		self, zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange,
	) -> None:
		#print("Service %s of type %s state changed: %s" % (name, service_type, state_change))

		if state_change is ServiceStateChange.Added:
			info = zeroconf.get_service_info(service_type, name)
			if (service_type == '_ravenna_session._sub._rtsp._tcp.local.'):
				name = name.split('.',1)[0]
				address = socket.inet_ntoa(cast(bytes, info.address))
				port = cast(int, info.port)
				req = 'DESCRIBE rtsp://' + address + ':' + str(port) + '/by-name/' + name + ' RTSP/1.0\r\n\r\n'
				s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
				s.connect((address, port))
				s.send(req.encode())
				sdp_text = s.recv(4096).decode()
				self.add_aoip_service_from_sdp_text("Ravenna", sdp_text)


if __name__ == '__main__':

	root = Tk()

	def quit_discovery():
		root.quit()

	def run_discovery():
		root.wm_title("Listening for Services")
		#root.pack_propagate(0)
		text_box = Text(root, width = 40, height = 25)
		text_box.pack()
		button = Button(root, text="Quit", command=quit_discovery)
		button.pack()

		def console_print(string):
			text_box.insert(END, string + '\n')
			text_box.see("end")

		def new_service_callback(service):
			console_print("----------------------------------------")
			console_print("Name: " + service.name)
			console_print("System: " + service.system)
			console_print("IP Address:" + service.ip_address)
			console_print("Port:" + str(service.port))
			console_print("Codec:" + service.codec)
			console_print("Channels:" + str(service.sdp.channels))
			console_print("Sampling Frequency:" + str(service.sdp.fs))
			#print("SDP:")
			#service.sdp.print()
			console_print("----------------------------------------")

		new_aoip_discovery = aoip_discovery(new_service_callback)

		root.mainloop()
		new_aoip_discovery.stop()
	
	run_discovery()








	


"""
 AOIP Services 
 Copyright (c) 2019, Dolby Laboratories Inc.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification, are permitted
 provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
    and the following disclaimer in the documentation and/or other materials provided with the distribution.
 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or
    promote products derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import socket
import struct
import sys
import netifaces
import platform

class MulticastGroup:

    interface = None
    sock = None
    address = None
    port = None
    localIp = None

    def join(self,interface, address, port):
        """
        Creates a socket, sets the necessary options on it, then binds it. The socket is then returned for use.
        """

        self.interface = interface
        self.localIp = netifaces.ifaddresses(interface)[netifaces.AF_INET][0]['addr']
        self.port = port
        self.address = address
        # create socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # Not all platforms support REUSEPORT i.e. Windows does not
        # The linux check below won't always work e.g. 3.10 will fail but it is close enough
        if platform.system() == 'Darwin' or (platform.system() == 'Linux' and platform.release() > "3.9"):
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(self.localIp))
        self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        membership_request = socket.inet_aton(self.address) + socket.inet_aton(self.localIp)

        # See http://www.tldp.org/HOWTO/Multicast-HOWTO-6.html for explanation of sockopts
        self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, membership_request)

        # Bind the socket to an interface depending on system
        if platform.system() == 'Darwin':
            self.sock.bind(('0.0.0.0', self.port))
        else:
            self.sock.bind((self.localIp, self.port))

    def leave(self):
        # leave group if active
        if (self.address != None):
            membership_request = socket.inet_aton(self.address) + socket.inet_aton(self.localIp)
            self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_DROP_MEMBERSHIP, membership_request)
            self.sock.close()
            self.address = None
            self.interface = None
            self.port = None
            self.localIp = None